#ifndef _CAMERA_H_
#define _CAMERA_H_
#include "WPILib.h"
#include "DSmap.h"

#define NO_CAMERA 1

class WeejusCamera {
protected:
	const static int MAX_FPS = 10;
	const static int BRIGHTNESS = 50;
	const static int COLOR_LEVEL = 50;
	const static int EXPOSURE_PRIORITY = 0;
	const static int COMPRESSION = 0;

	const static AxisCameraParams::Resolution_t RESOLUTION =
			AxisCameraParams::kResolution_640x480;
	const static AxisCameraParams::Exposure_t EXPOSURE_CONTROL =
			AxisCameraParams::kExposure_Automatic;
	const static AxisCameraParams::WhiteBalance_t WHITE_BALANCE =
			AxisCameraParams::kWhiteBalance_Automatic;

	const static float SERVO_LOCAL_DELTA_ANGLE = 140.0;

	const static float SERVO_LOCAL_UP_ANGLE = 180.0;
	const static float SERVO_LOCAL_DOWN_ANGLE = 0.0;
	const static float SERVO_RANGE = 180.0;

public:
	WeejusCamera();
	//PLEASE DO NOT CALL THIS
	~WeejusCamera();

	/*
	 * Method: setLight(bool)
	 * 
	 * Inputs:
	 * 			bool on: value representing whether to turn the light on or off
	 * Outputs: Null
	 * 
	 * Remarks:
	 * controls the ring light
	 * light will stay on until call again with false
	 */
	void setLight(bool on);

	/*
	 * Method: flip(bool);
	 * 
	 * Inputs:
	 * 			bool flipped: represents whether to flip the camera
	 * Outputs: Null
	 * 
	 * Remarks:
	 * tells the camera to flip orientations and reverse the video feed
	 */
	void flip(bool flipped);

	/*
	 * Method: tiltUp(float)
	 * 
	 * Inputs:
	 * 			float amount: degrees to tilt the
	 * 				 camera in degrees (range -90
	 * 				 -> 90) relative to current 
	 * 				position
	 * Outputs: Null
	 * 
	 * Remarks:
	 * tilts the camera upward by the specified degree
	 *  with relation to the current view
	 * if given amount will move camera out of range,
	 *  amount will be changed so it does not
	 */
	void tiltUp(float amount);

	/*
	 * Method: tiltDown(float)
	 * 
	 * Inputs:
	 * 			float amount: degrees to tilt the
	 * 				 camera in degrees (range -90
	 * 				 -> 90) relative to current 
	 * 				position
	 * Outputs: Null
	 * 
	 * Remarks:
	 * tilts the camera downward by the specified degree
	 *  with relation to the current view
	 * if given amount will move camera out of range,
	 *  amount will be changed so it does not
	 */
	void tiltDown(float amount);

	/*
	 * Method: getAngle();
	 * 
	 * Inputs: Null
	 * Outputs: float representing the current angle
	 *     of the camera in degrees
	 * 
	 * Remarks:
	 * 0 degrees is horizontal (pointing straight
	 *  forward)
	 */
	float getAngle();

protected:
	/*
	 * Method: getImage()
	 * 
	 * Inputs: Null
	 * Outputs: image retrieved from the camera
	 * 
	 * Remarks:
	 * returns NULL if the camera did not send
	 *   a fresh image
	 * RECIEVER MUST DELETE THE IMAGE
	 */
	HSLImage* getImage();

	/*
	 * Method: getDebugLine(int);
	 * 
	 * Inputs:
	 * 			int line: integer representing the debug line to get (range 1->6)
	 * Outputs: char array containing the debug line to print
	 * 
	 * Remarks:
	 * will return the line specified by the number passed to the parameter
	 * only can return 6 debug lines
	 * if line < 1, will return line 1
	 * if line > 6, will return line 6
	 */
	char* getDebugLineCamera(int line);
private:
	float localTilt(float deg, float min, float max);

#if !NO_CAMERA
	AxisCamera& m_camera;
#endif

	// TODO change type to PWM, subclass?
	Servo m_tiltSrv;
	Relay m_lightRing;
	float m_servo;
	float m_servoShooter;
	float m_servoClimber;

	bool m_globalFlipped;

	// buffer to contain debug line sent to controls
	char _dbgBuf[DEBUG_LINE_LENGTH];
};

#endif

