#include "WeejusClimber.h"

// debug gunk;
#include "WeejusConfig.h"

WeejusClimber* WeejusClimber::theClimber = NULL;

WeejusClimber* WeejusClimber::getInstance() {
	if (theClimber == NULL) {
		theClimber = new WeejusClimber();
	}
	return theClimber;
}

WeejusClimber::WeejusClimber() {

	// create sensors
	climb_left_sensor = new DigitalInput(DIG_CLIMB_HOOK_LEFT_SIDE_ENGAGED);
	climb_right_sensor = new DigitalInput(DIG_CLIMB_HOOK_RIGHT_SIDE_ENGAGED);

	fixed_lh_arm_sensor = new DigitalInput(DIG_FIXED_CLIMB_LH_ENGAGED);
	fixed_rh_arm_sensor = new DigitalInput(DIG_FIXED_CLIMB_RH_ENGAGED);

	//alignment_sensor = new AnalogChannel(ANA_ROBOT_VEE_ALIGN_SENSOR);
	deploy_servo = new Servo(PWM_CLIMBER_DEPLOY_RH_SERVO);
	dumper_servo = new Servo(PWM_DUMPER_RELEASE_SERVO);

	// create Climb drive
	arm_drive = new WeejusClimberDrive();

	// get access to driver and shooter
	theDrive = WeejusDrive::getInstance();
	theShooter = WeejusShooter::getInstance();

	// broken switches
	limits_broken = false;
#if FORCE_BROKEN_AUTO_CLIMB
	auto_broken = true;
#else
	auto_broken = false;
#endif
	climber_broken = false;

	// this does not need to be set;
	// initialize does it as well, but this
	// can't hurt
#if FORCE_BROKEN_AUTO_CLIMB
	// arm_drive is enabled by default
	climber_state = MANUAL;
#else
	climber_state = AUTO;
#endif

	initialize();
}

// inits state variables
void WeejusClimber::initialize() {
	arm_drive->initialize();

	// these states are initialized in here,
	// so that state always corresponds to 
	// the normal; without this, the
	// climb_begun AUTO-entry lock would make
	// state changes ugly
	if (climber_broken) {
		climber_state = NUKED;
	} else {
		if (auto_broken) {
			climber_state = MANUAL;
		} else {
			climber_state = AUTO;
		}
	}
	DBG_PRINTF(DBG_CLIMBER_MANUAL, "[Climber] at initialization, state set to %s\n", ModeToString(climber_state));

	//
	// Only in AUTO, we start with arms
	// disabled so that they are only
	// enabled when the climb button is
	// held down
	//
	if (climber_state == AUTO) {
		arm_drive->disable();
	}

	deploy_servo->Set(PIN_REST);
	dumper_servo->Set(DUMPER_REST);

	ac_state = INITIAL;

	climb_begun = false;
	deployed = false;
	dumped = false;

	// both arms on the same bar; that is, none ;-)
	// the value doesn't matter
	lift_state = PULLING;

	climb_active = false;
	climb_failed = false;
}

WeejusClimber::~WeejusClimber() {
	// delete all objects we create
	delete arm_drive;

	//delete alignment_sensor;

	delete climb_left_sensor;
	delete climb_right_sensor;
	delete fixed_lh_arm_sensor;
	delete fixed_rh_arm_sensor;

	delete deploy_servo;
}

// Controls deals with all the gory details of who 
// controls the drive. 
void WeejusClimber::processAlign() {
	//
	// Forward is defined are relative to the robot;
	// in aligning, we are moving backwards into the pyramid
	//
	//             F        |
	//          +-----+     | 
	//       L  |  _  |  R  |
	//          +-/.\-+     |
	//          I /|\ I     |
	//           / | \      |
	//          /  .  \     |
	//         /  /|\  \    |
	//
	// The "I"s are arms; The left arm is on the left side;
	// the robot moves backward toward the pyramid

	bool ls, rs, e;
	e = getEngaged();
	ls = getLeftArmSensor();
	rs = getRightArmSensor();

	if (ls && rs) {
		// allow climbing
		DBG_PRINTF(DBG_CLIMBER_AUTO, "[Climber] Arms engaged; Proceeding\n");
		return;
	}

	// left side engaged
	if (ls) {
		// turn left, so right side moves in
		DBG_PRINTF(DBG_CLIMBER_AUTO, "[Climber] Turning left back to align\n");
		theDrive->pyramidAlignmentMotion(WeejusDrive::LeftBackTurn);
	} else if (rs) {// right side
		// turn right, so left side moves in
		DBG_PRINTF(DBG_CLIMBER_AUTO, "[Climber] Turning right back to align\n");
		theDrive->pyramidAlignmentMotion(WeejusDrive::RightBackTurn);
	} else if (e) {// no arms 
		// move back; neither sensor touches
		DBG_PRINTF(DBG_CLIMBER_AUTO, "[Climber] Going backward to align\n");
		theDrive->pyramidAlignmentMotion(WeejusDrive::Backward);
	} else {// no connection to pyramid
		ac_state = DEPLOY;
		DBG_PRINTF(DBG_CLIMBER_AUTO, "[Climber] Stopping alignment; disconnect from pyramid\n");
		theDrive->stopAction();
	}
	return;

}

// this is only called when "climb_active" true
void WeejusClimber::processAuto() {
	switch (ac_state) {
	case INITIAL:
		break;
	case DEPLOY:
		if (getEngaged() && arm_drive->atMiddle()) {
			DBG_PRINTF(DBG_CLIMBER_AUTO,
					"[Climber] Engaged and arm at middle; Switching into aligning state\n");
			ac_state = ALIGNING;
		}
		break;
	case ALIGNING:
		// transition to LEVEL_1
		// occurs when one is aligned and has pressed
		// the start button

#if !SKIP_ALIGN
		processAlign();
#endif			

		if (getAligned()) {
#if !SKIP_ALIGN
			// set drives to slow forward power - keep it aligned
			// so it doesn't mechanically pull back
			DBG_PRINTF(DBG_CLIMBER_AUTO,
					"[Climber] Creeping forward\n");
			theDrive->pyramidAlignmentMotion(WeejusDrive::BackwardCreep);
#endif
			// lower the arms and pull self up
			DBG_PRINTF(DBG_CLIMBER_AUTO,
					"[Climber] Begin Downstroke; climb_begun set to true;\n");
			downStroke();
			climb_begun = true;
			ac_state = LEVEL_1;

			DBG_PRINTF(DBG_CLIMBER_AUTO, "[Climber] entering PULLING state\n");
			lift_state = PULLING;
		}
		break;

		// move up; move down
	case LEVEL_1:
	case LEVEL_2:
		// when hooks tripped, move down
		switch (lift_state) {

		case PULLING:// shuttle down
			if (reachedDownStroke()) {
#if !SKIP_ALIGN
				if (ac_state == LEVEL_1) {
					// stop moving the drive as soon as we have contact;
					// would we stop it earlier, we might lose the

					DBG_PRINTF(DBG_CLIMBER_AUTO,
							"[Climber] Stopped drive;\n");
					theDrive->stopAction();
				}
#endif				
				// go back up again
				DBG_PRINTF(DBG_CLIMBER_AUTO, "[Climber] entering RAISING state\n");
				lift_state = WAITING_FOR_ARMS;
				DBG_PRINTF(DBG_CLIMBER_AUTO,
						"[Climber] Upstroke\n");
				upStroke();
			} else if (failedDownStroke()) {
				DBG_PRINTF(DBG_CLIMBER_AUTO, "[Climber] hit bottom limit\n");
				failClimb();
			} else if (!reachedUpStroke()) {
				DBG_PRINTF(DBG_CLIMBER_AUTO, "[Climber] top arms disconnect\n");
				failClimb();
			}
			break;

		case WAITING_FOR_ARMS:
			if (!reachedUpStroke()) {// must wait for climbing arms to detach from lower rung
				DBG_PRINTF(DBG_CLIMBER_AUTO, "[Climber] stopping the wait for the ARMS arm disconnect\n");
				lift_state = RAISING;
			} else if (failedUpStroke()) {
				failClimb();
			}

			break;

		case RAISING:// shuttle up
			if (reachedUpStroke()) {
				DBG_PRINTF(DBG_CLIMBER_AUTO, "[Climber] entering PULLING state\n");
				lift_state = WAITING_FOR_FIXEDS;

				DBG_PRINTF(DBG_CLIMBER_AUTO,
						"[Climber] Downstroke;\n");
				downStroke();

				if (ac_state == LEVEL_1) {
					ac_state = LEVEL_2;
				} else if (ac_state == LEVEL_2) {
					ac_state = LEVEL_3;
				}
			} else if (failedUpStroke()) {
				DBG_PRINTF(DBG_CLIMBER_AUTO, "[Climber] hit top limit\n");
				failClimb();
			} else if (!reachedDownStroke()) {
				DBG_PRINTF(DBG_CLIMBER_AUTO, "[Climber]  bottom arms disconnect\n");
				failClimb();
			}
			break;

		case WAITING_FOR_FIXEDS:
			if (!reachedDownStroke()) {// must wait for climbing arms to detach from lower rung
				DBG_PRINTF(DBG_CLIMBER_AUTO, "[Climber] stopping the wait for the FIXED arm disconnect\n");
				lift_state = PULLING;
			} else if (failedDownStroke()) {
				failClimb();
			}

			break;
		}

		break;
	case LEVEL_3:
		if (lift_state == WAITING_FOR_FIXEDS) {
			if (!reachedDownStroke()) {// must wait for climbing arms to detach from lower rung
				lift_state = PULLING;
			} else if (failedDownStroke()) {
				failClimb();
			}
		} else if (lift_state == PULLING) {

			if (reachedDownStroke()) {
				DBG_PRINTF(DBG_CLIMBER_AUTO,
						"[Climber] Finished; Stopping;\n");
				arm_drive->disable();
				ac_state = DONE;
			} else if (failedDownStroke() || !reachedUpStroke()) {
				failClimb();
			}
		}
		break;
	case DONE:
		break;

	default:
		break;
	}
}

void WeejusClimber::failClimb() {
	DBG_PRINTF(DBG_CLIMBER_AUTO,
			"[Climber] Climb failed; entering manual mode;\n");
	climber_state = MANUAL;
	climb_failed = true;
}

bool WeejusClimber::failedClimb() {
	return climb_failed;
}

bool WeejusClimber::reachedUpStroke() {
	return (getLeftArmSensor() && getRightArmSensor());
}

// the up strokes fail:
// when the arms reach the top shuttle limit
// when any of the lower arms are not on the lower rung
bool WeejusClimber::failedUpStroke() {
	return (arm_drive->wasTopLimitHit());
}

bool WeejusClimber::reachedDownStroke() {
	return getLeftFixedSensor() && getRightFixedSensor();
}

// the down strokes fail
// when the arms reach the bottom limit without connecting
bool WeejusClimber::failedDownStroke() {
	return (arm_drive->wasBottomLimitHit());
}

void WeejusClimber::downStroke() {
	arm_drive->setPower(CLIMB_DRIVE_DOWN_POWER);
}

void WeejusClimber::upStroke() {
	arm_drive->setPower(CLIMB_DRIVE_UP_POWER);
}

void WeejusClimber::process() {
	if (climber_state == AUTO) {
		// the climb failed check may not be needed
		if (climb_active && !climb_failed) {
			processAuto();
		}
	}

	arm_drive->process();
}

bool WeejusClimber::deployArms() {
	DBG_PRINTF(DBG_CLIMBER_MANUAL, "[Climber] Trying to deploy\n");
	if (climber_state == NUKED) {
		DBG_PRINTF(DBG_CLIMBER_MANUAL,
				"[Climber] Not deploying; Nuked state\n");
		return false;
	}

	if (deployed) {
		DBG_PRINTF(DBG_CLIMBER_MANUAL,
				"[Climber] Not deploying; already did\n");
		return false;
	}

	if (climber_state == MANUAL || (climber_state == AUTO && ac_state
			== INITIAL)) {
		// only deploy if the climber is in stored position
		if (theShooter->getClimbReady()) {
			DBG_PRINTF(DBG_CLIMBER_MANUAL, "[Climber] Deploying arms\n");
			ac_state = DEPLOY;
			deployed = true;
			if (climber_state == AUTO) {
				arm_drive->enable();
				DBG_PRINTF(DBG_CLIMBER_AUTO, "[Climber] Moving shuttle to center pos\n");
				arm_drive->goToMiddle();
			}
			PullPins();
		} else {
			DBG_PRINTF(DBG_CLIMBER_MANUAL, "[Climber] Shooter not in proper position for deployment\n");
			return false;
		}
	} else {
		DBG_PRINTF(DBG_CLIMBER_MANUAL, "[Climber] Failed state check for deployment\n");
	}

	return true;
}

bool WeejusClimber::dump() {
	if (dumped) {
		DBG_PRINTF(DBG_CLIMBER_MANUAL, "[Climber] Already dumped; aborting\n");
		return false;
	}

	DBG_PRINTF(DBG_CLIMBER_MANUAL, "[Climber] Dumping discs\n");

	dumper_servo->Set(DUMPER_DEPLOYED);
	dumped = true;

	return true;
}

bool WeejusClimber::getEngaged() {
	// TODO add V sensor when it is ready
	return true;
}

void WeejusClimber::startClimb() {
	climb_active = true;
	if (climber_state != AUTO) {
		return;
	}

	// stop drive after control is handed over to us by Controls
	theDrive->setMotorSpeed(0, 0);
	// setup or resume
	arm_drive->enable();
}

void WeejusClimber::stopClimb() {
	climb_active = false;
	// you can only stop in auto_mode
	if (climber_state != AUTO) {
		return;
	}

#if !SKIP_ALIGN
	if (ac_state == ALIGNING) {
		// turn down the drive
		theDrive->stopAction();
	}
#endif

	arm_drive->disable();
}

// asks for alignment
bool WeejusClimber::getAligned() {
	// identical to reached_Up_Stroke
	return (getLeftArmSensor() && getRightArmSensor());
}

/*
 * Which tower level the bot is at
 */
int WeejusClimber::getLevel() {
	switch (ac_state) {
	case LEVEL_2:
		return 1;
	case LEVEL_3:
		return 2;
	case DONE:
		return 3;
	default:
		return 0;
	}
}

bool WeejusClimber::getLevelAchieved() {
	return reachedDownStroke();
}

// BAD-GOOD only in non-climbing, enforce automatic control
// GOOD->BAD disable drive, go to manual control
// returns true if the action was successful
//
bool WeejusClimber::setIsAutomatic(bool automatic) {
#if FORCE_BROKEN_AUTO_CLIMB
	auto_broken = true;
	return true;
#else
	auto_broken = !automatic;

	if (climber_state == NUKED) {
		DBG_PRINTF(DBG_CLIMBER_MANUAL, "[Climber] In nuked mode; not setting AUTO or MANUAL\n");
		return false;
	}

	if (auto_broken) {
		DBG_PRINTF(DBG_CLIMBER_MANUAL, "[Climber] Climber state is now MANUAL\n");
		climber_state = MANUAL;
		arm_drive->enable();
#if !SKIP_ALIGN
		if (ac_state == ALIGNING) {
			theDrive->stopAction();
		}
#endif
		return true;
	} else if (!climb_begun) {
		DBG_PRINTF(DBG_CLIMBER_MANUAL, "[Climber] Climber state is now AUTOMATIC\n");
		climber_state = AUTO;
		// disable so as to ensure that
		// the climber is inactive
		// until the climb button is pressed
		arm_drive->disable();
		return true;
	} else {
		DBG_PRINTF(DBG_CLIMBER_MANUAL, "[Climber] Climb has begun; remaining in MANUAL\n");
		return false;
	}
#endif
}

void WeejusClimber::setDead(bool dead) {
	DBG_PRINTF(DBG_CLIMBER_MANUAL, "[Climber] Setting Dead State to %d\n", dead);

	climber_broken = dead;

	// an explicit decision was made to not
	// allow a climber to revert to "undead" if
	// a climb was already started (unsafe)
	if (dead) {
		DBG_PRINTF(DBG_CLIMBER_MANUAL, "[Climber] Setting to NUKED\n");
#if !SKIP_ALIGN
		if (climber_state == AUTO && ac_state == ALIGNING) {
			theDrive->stopAction();
		}
#endif
		climber_state = NUKED;
		arm_drive->disable();
	} else {
		if (auto_broken) {
			DBG_PRINTF(DBG_CLIMBER_MANUAL, "[Climber] Setting to MANUAL\n");
			climber_state = MANUAL;
		} else if (!climb_begun) {
			DBG_PRINTF(DBG_CLIMBER_MANUAL, "[Climber] Setting to AUTO\n");
			climber_state = AUTO;
		} else {
			DBG_PRINTF(DBG_CLIMBER_MANUAL, "[Climber] Setting to MANUAL\n");
			climber_state = MANUAL;
		}

		arm_drive->enable();
	}
}

void WeejusClimber::setLimitsBroken(bool broken) {
	limits_broken = broken;
	DBG_PRINTF(DBG_CLIMBER_MANUAL, "[Climber] Setting pot broken to %d\n", limits_broken);
	arm_drive->setLimitsBroken(broken);
	return;
}

void WeejusClimber::moveArms(float power, bool override) {
	if (climber_state == MANUAL) {
		// If the override button has not been
		// pressed before deployment, cut
		// the power
		if (!deployed && !override) {
			arm_drive->setPower(0.0f);
			return;
		}

		// do not set climb to begun if
		// nothing has been done with the
		// arms: i.e., null power set all
		// the time
		if (!climb_begun && power != 0.0f) {
			climb_begun = true;
		}
		arm_drive->setPower(power);
	}
	return;
}

void WeejusClimber::PullPins() {
	DBG_PRINTF(DBG_CLIMBER_MANUAL, "[Climber] Servo to %f; pulling pins\n", PIN_PULLED);
	deploy_servo->Set(PIN_PULLED);
}

// this sensor is tripped at low
bool WeejusClimber::getLeftArmSensor() {
	return !climb_left_sensor->Get();
}
// this sensor is tripped at low
bool WeejusClimber::getRightArmSensor() {
	return !climb_right_sensor->Get();
}

// this sensor is tripped at high
bool WeejusClimber::getLeftFixedSensor() {
	return fixed_lh_arm_sensor->Get();
}
// this sensor is tripped at high
bool WeejusClimber::getRightFixedSensor() {
	return fixed_rh_arm_sensor->Get();
}

char* WeejusClimber::getDebugLine(int line) {
	switch (line) {
	case 1:
		snprintf(_dbgBuf, sizeof(_dbgBuf), "CLB L: %i R: %i",
				getLeftArmSensor(), getRightArmSensor());
		break;
	case 2:
		snprintf(_dbgBuf, sizeof(_dbgBuf), "HI: %i FIXL: %i FIXR: %i",
				arm_drive->getTopLimit(), getLeftFixedSensor(),
				getRightFixedSensor());
		break;
	case 3:
		snprintf(_dbgBuf, sizeof(_dbgBuf), "Lo: %i VEE: %i",
				arm_drive->getBottomLimit(), getEngaged());
		break;
	case 4:
		snprintf(_dbgBuf, sizeof(_dbgBuf), "MID: %i CURRENT: %.4f",
				arm_drive->getMiddleLimit(),
				(double) arm_drive->getAvgCurrent());
		break;
	case 5:
		snprintf(_dbgBuf, sizeof(_dbgBuf), "POT: %.4f VEL: %.6f",
				arm_drive->getPos(), arm_drive->getSpeed());
		break;
	case 6:
		snprintf(_dbgBuf, sizeof(_dbgBuf), "M %s|AS %s|LS %s",
				ModeToString(climber_state), StateToString(ac_state),
				LiftStateToString(lift_state));
		break;
	default:
		snprintf(_dbgBuf, sizeof(_dbgBuf), "null ");
		break;
	}
	return _dbgBuf;
}
char* WeejusClimber::StateToString(WeejusClimber::autoclimb_state state) {
	switch (state) {
	case WeejusClimber::INITIAL:
		return "INIT";
	case WeejusClimber::DEPLOY:
		return "DEPL";
	case WeejusClimber::ALIGNING:
		return "ALIG";
	case WeejusClimber::LEVEL_1:
		return "LEV1";
	case WeejusClimber::LEVEL_2:
		return "LEV2";
	case WeejusClimber::LEVEL_3:
		return "LEV3";
	case WeejusClimber::DONE:
		return "DONE";
	}
	return "";
}

char* WeejusClimber::ModeToString(WeejusClimber::climb_mode mode) {
	switch (mode) {
	case WeejusClimber::AUTO:
		return "AUTO";
	case WeejusClimber::MANUAL:
		return "MAN";
	case WeejusClimber::NUKED:
		return "NUKD";
	}
	return "";
}

char* WeejusClimber::LiftStateToString(WeejusClimber::ac_lift_state lift_state) {
	switch (lift_state) {
	case WeejusClimber::RAISING:
		return "RAISE";
	case WeejusClimber::PULLING:
		return "PULL";
	case WeejusClimber::WAITING_FOR_ARMS:
		return "WARMS";
	case WeejusClimber::WAITING_FOR_FIXEDS:
		return "WFXDS";
	}
	return "";
}
