#ifndef _CLIMBER_H_
#define _CLIMBER_H_
#include "WPILib.h"

#include "WeejusClimberDrive.h"
#include "WeejusShooter.h"
#include "WeejusDrive.h"
#include "DSmap.h"
#include "UDPLogger.h"

#define FORCE_BROKEN_AUTO_CLIMB 1

/*
 Sensors:
 
 one on tip of each arm of know that arm
 has engaged the pole (4 total)
 
 one on "V" to know when it has engaged the
 pyramid (climbing cannot begin until engaged)
 one on each arm at height of horizontal bar
 (feedback to allow for alignment) (2 total)
 
 
 |encoder for motor shaft (keep speed constant
 b/c varying loads)
 upper and lower limit sensors on arm track
 
 
 Actuators:
 servo(s) to release arms for deployment
 2 motors & jaguars
 
 
 Lights/Indicators:
 "v" has engaged pyramid
 arms are aligned & in contact with
 pyramid (i.e. complete alignment)
 level acheived!

 Dashboard:
 level indicator
 extending arms are latched on pyramid
 
 In other words, there are five degrees of kaputness:
 perfect
 broken pot
 broken autoclimb
 broken autoclimb & pot
 broken mechanism
 */

// up; .5
// down; -.8
// test; (+-).3;
#define	CLIMB_DRIVE_UP_POWER	 	(0.3)	// climber power for up movement if climber in manual mode
#define	CLIMB_DRIVE_DOWN_POWER		(-0.3)	// climber power for dn movement if climber in manual mode
#define PIN_REST        (0.7)
#define PIN_PULLED      (0.1)

#define	DUMPER_REST		(0.34)
#define DUMPER_DEPLOYED	(0.75)

// AUTO skips the alignment sections
#define SKIP_ALIGN 1

class WeejusClimber {

public:

	/*
	 * Method: getInstance();
	 * 
	 * Inputs: Null
	 * Outputs: pointer to the only WeejusClimber object
	 * 
	 * Remarks:
	 * Returns a pointer to the only instance of the
	 *  WeejusClimber. If the instance has not been
	 *  created, it will be, otherwise the existing
	 *  instance will be returned
	 *
	 */
	static WeejusClimber* getInstance();

	/*
	 * Method: initialize()
	 * Inputs: Null
	 * Outputs: Null
	 * Remarks: this resets all state variables
	 * 			 to what they should be at the
	 * 			 start of a match; It does NOT 
	 * 			set broken switches
	 * 
	 */
	void initialize();

	/*
	 * Method: deployArms();
	 * 
	 * Inputs: Null
	 * Outputs: bool indicating if the arms were deployed
	 * 				true if arms deployed or already deployed
	 * 				false could not deploy arms
	 *                (i.e. not at angle)
	 * 
	 * Remarks: 
	 * Deploys the arms to the tower
	 * Called by WeejusControls when operator
	 *  presses deploy button
	 */
	bool deployArms();

	/*
	 * Method: dump();
	 * 
	 * Inputs: Null
	 * Outputs: bool indicating if process of dumping succeeded (true)
	 * 				 or had happened or failed (false)
	 * 
	 * Turns the dumper servo, to release disks onto pyramid
	 * Called by WeejusControls when operator presses
	 *  DUMPER_DEPLOY button
	 */
	bool dump();

	/*
	 * Method: getEngaged();
	 * 
	 * Inputs: Null
	 * Outputs: bool representing whether the "V" mechanism has
	 *           engaged the pyramid
	 * 
	 * Remarks: 
	 * Returns whether the "V" has engaged the pyramid or not
	 * Called by WeejusControls to turn on indicator light for driver feedback
	 * 
	 * don't call until 1 or 2 years of usage (18+)
	 */
	bool getEngaged();

	/*
	 * Method: startClimb();
	 * 
	 * Inputs: Null
	 * Outputs: Null
	 * 
	 * Remarks: 
	 * Starts "auto" climbing (call once to start, NOT once per loop to continue)
	 * Called by WeejusControls when drivers want robot to start climbing
	 * Once called robot will handle all climbing actions
	 * Automatically determines where the robot is and starts from that point (i.e. needs to align, already on pyramid, etc)
	 * 
	 * Does not work if broken is set
	 * 
	 * Typically begins by auto-aligning, then ascends the pyramid
	 */
	void startClimb();

	/*
	 * Method: stopClimb();
	 * 
	 * Inputs: Null
	 * Outputs: Null
	 * 
	 * Remarks:
	 * Will stop all actions of the climber (call once to stop, likely when climb button released)
	 * If needed, call startClimb() to resume from where left off
	 * Does not work if broken is set
	 */
	void stopClimb();

	/*
	 * Method: getAligned()
	 * 
	 * Inputs: Null
	 * Outputs: bool representing whether the robot is aligned with the pyramid
	 * 
	 * Remarks:
	 * Returns whether the robot is aligned and that it is safe to begin climbing
	 * Called by WeejusControls to turn on indicators for driver feedback
	 */
	bool getAligned();

	/*
	 * Method: getLevel();
	 * 
	 * Inputs: Null
	 * Outputs: integer representing what level the robot thinks it's at (or passed)
	 * 
	 * Remarks:
	 * Returns what level the robot thinks it's at
	 * Called by WeejusControls to provide feedback to the drivers
	 */
	int getLevel();

	/*
	 * Method: getLevelAchieved();
	 * 
	 * Inputs: Null
	 * Outputs: bool representing whether robot
	 *  is on a current level, or in transition
	 * 
	 * Remarks: Null
	 */
	bool getLevelAchieved();

	/*
	 * Method: setIsAutomatic(bool);
	 * 
	 * Inputs:
	 * 			bool auto: value representing whether the climber can be operated automatically
	 * Outputs: if the change succeeded
	 * 
	 * Remarks:
	 * Tells WeejusClimber that 1+ arm sensor(s) are broken, meaning all movement should be manual
	 * Called by WeejusControls when a broken switch is flipped
	 */
	bool setIsAutomatic(bool automatic);

	/*
	 * Method: setDead();
	 * 
	 * Inputs:
	 * 			bool dead: value representing whether the climber should not do anything
	 * Outputs: Null
	 * 
	 * Remarks:
	 * When true, neither automatic nor manual movement nor deployment will work
	 * Called by WeejusControls when a broken switch is flipped, saying the climber is unsafe to operate
	 * Sets the limits broken
	 */
	void setDead(bool dead);

	/*
	 * Method: setLimitsBroken(bool);
	 * 
	 * Inputs:
	 * 			bool broken: value representing whether the pot on
	 * 				 the "winch" is broken or not
	 * Outputs: Null
	 * 
	 * Remarks:
	 * sets pot as broken (maybe use PID)
	 */
	void setLimitsBroken(bool broken);

	/*
	 * Method: moveArms(float);
	 * 
	 * Inputs:
	 * 			float power: value representing power to send to BOTH climber motors (range -1.0 -> 1.0)
	 * 			bool overide: climber always moves during MANUAL
	 * Outputs: Null
	 * 
	 * Remarks:
	 * Tells WeejusClimber to move arms in a certain direction at the specified power
	 * negative value = move shuttle down
	 * positive value = move shuttle up
	 * Called by WeejusControls when drivers need to manually operate the climber
	 * Auto Broken MUST be set in order for this to work
	 * NOT recommended to use this method if sensors are not broken
	 */
	void moveArms(float power, bool override);

	/*
	 * Method: failedClimb()
	 * 
	 * Inputs: Null
	 * Outputs: Null
	 * 
	 * Remarks: to be called by Lights when
	 * a climb fails
	 * 
	 * 
	 */
	bool failedClimb();

	/*
	 * Method: process();
	 * 
	 * Inputs: Null
	 * Outputs: Null
	 * 
	 * Remarks:
	 * Updates all motors and internal status variables
	 * MUST be called at the END of each loop to continue operation of the climber
	 */
	void process();

	/*
	 * Method: getDebugLine(int);
	 * 
	 * Inputs:
	 * 			int line: integer representing the debug line to get (range 1->6)
	 * Outputs: char array containing the debug line to print
	 * 
	 * Remarks:
	 * will return the line specified by the number passed to the parameter
	 * only can return 6 debug lines
	 */
	char* getDebugLine(int line);
private:
	// or just "enum {---} climber_state;"
	typedef enum {
		AUTO, MANUAL, NUKED
	} climb_mode;
	typedef enum {
		INITIAL, DEPLOY, ALIGNING, LEVEL_1, LEVEL_2, LEVEL_3, DONE
	} autoclimb_state;

	typedef enum {
		RAISING, // top arms moving up; bottom arms latched
		PULLING, // top arms fixed; robot rising
		WAITING_FOR_ARMS,
		WAITING_FOR_FIXEDS,
	} ac_lift_state;

	/*
	 * 
	 * Constructs the Climber:
	 * Broken switches are set to FALSE here.
	 */
	WeejusClimber();
	~WeejusClimber();

	void processAuto();
	void processAlign();

	/* set things to pull arms down, robot up*/
	void downStroke();
	void upStroke();
	bool reachedDownStroke();
	bool reachedUpStroke();
	bool failedUpStroke();
	bool failedDownStroke();

	bool getLeftArmSensor();
	bool getRightArmSensor();
	bool getLeftFixedSensor();
	bool getRightFixedSensor();

	void PullPins();

	char* StateToString(WeejusClimber::autoclimb_state state);
	char* ModeToString(WeejusClimber::climb_mode mode);
	char* LiftStateToString(WeejusClimber::ac_lift_state lift_state);

	/* switch to manual mode, set a flag on climb failure*/
	void failClimb();

	static WeejusClimber* theClimber;

	WeejusDrive* theDrive;
	WeejusShooter* theShooter;

	WeejusClimberDrive* arm_drive;

	// sensor digital input
	DigitalInput * climb_left_sensor;
	DigitalInput * climb_right_sensor;

	DigitalInput * fixed_lh_arm_sensor;
	DigitalInput * fixed_rh_arm_sensor;

	// sensor analog input
	//AnalogChannel * alignment_sensor; 

	// PWM outputs
	Servo * deploy_servo;
	Servo * dumper_servo;

	climb_mode climber_state;
	autoclimb_state ac_state;

	// Have the arms been moved so that the robot
	//  may no longer be on the floor.
	// Set in auto and manual when the arms are first moved down
	bool climb_begun;

	// Arms have been deployed; pins pulled
	bool deployed;
	bool dumped;

	// the state of the arms during a
	// level increase; AUTO only
	ac_lift_state lift_state;

	// flag set by startClimb and stopClimb;
	// this being true means that the climbing procedure is being
	// enacted; motors are moving. When this is false; the procedure
	// should be paused
	bool climb_active;

	// Flag for the enjoyment of lights,
	//  to mark if the climbing attempt failed
	// Set when the arms reach a limit without connecting
	bool climb_failed;

	// reprs pot broken broken switch
	bool limits_broken;

	// marks the autonomous method as broken; we use manual mode
	bool auto_broken;

	// marks the entire climber as broken;
	bool climber_broken;

	// buffer to contain debug line sent to controls
	char _dbgBuf[DEBUG_LINE_LENGTH];
};

#endif
