#ifndef WEEJUSCLIMBERDRIVE_H_
#define WEEJUSCLIMBERDRIVE_H_

#include "WPILib.h"
#include "WeejusCANJaguar.h"
#include "WeejusPotEncoder.h"

/*
 * The purpose of this class is to safely encapsulate
 * the climber arm motion system. Everything to deal with
 * safely moving the arms up and down.
 * This class exists purely for the consumption of the WeejusClimber.
 * Transitions between pottyness and power
 *  should be clean
 * 
 * Someday: Calibration method for the pot,
 *  with values returned to smart dashboard
 */
class WeejusClimberDrive {
private:
	//
	// If this flag is set to true, the pot position
	// is used to aid the limit switches in controlling
	// the position of the shuttle; When set to false,
	// the potentiometer is only used for speed determination;
	//
	// Use this @false when you are in a hurry
	static const bool USING_POT_LIMITS = false;
public:
	typedef enum {
		BLIND = 0, BAD_POT = 1, BAD_LIMIT = 2, FULL = 3
	} action_mode;

	/*
	 * Method: ClimberDrive();
	 * 
	 * Inputs: None
	 * 
	 * Remarks: Broken switch values set to FALSE by default
	 * 
	 */
	WeejusClimberDrive();
	~WeejusClimberDrive();

	/*
	 * Resets all state variables to teleop
	 * start. Does NOT reset broken switch variables
	 */
	void initialize();

	/*
	 * Marks the pot as broken;
	 * Call on change
	 */
	void setLimitsBroken(bool broken);

	/*
	 * moves the arms to the middle position.
	 * automatically finishes
	 * 
	 */
	void goToMiddle();
	/*
	 *  did the above finish?
	 * If everything is broken, this returns TRUE
	 */
	bool atMiddle();

	/*
	 * Controls power output;
	 */
	void setPower(float power);
	float getPower();

	/*
	 * Turns off the drive system temporarily
	 * Does nothing in Blind mode
	 */
	void disable();

	/*
	 * Enables the drive system and resumes
	 *  previous actions
	 * Does nothing in BLIND mode
	 */
	void enable();

	/*
	 *  If the top or bottom
	 * limit switch was hit 
	 * SINCE THE LAST SETTING
	 * of mode or velocity/power
	 * 
	 */
	bool wasTopLimitHit();
	bool wasBottomLimitHit();

	/*
	 * Get the value of the limit sensors
	 */
	bool getTopLimit();
	bool getMiddleLimit();
	bool getBottomLimit();

	/*
	 * Returns the position of the climber vertically,
	 *  in volts
	 * 
	 */
	double getPos();
	/*
	 * Get the current from the motor
	 */
	float getAvgCurrent();
	/*
	 *  Get the speed from the pot
	 */
	double getSpeed();

	/*
	 * Method: process()
	 * Inputs: Null
	 * Outputs: Null
	 * 
	 * Update the drive system's state, dealing
	 * with the motor control and safety mechanisms
	 * (i.e, preventing the system from destroying itself.
	 * 
	 */
	void process();
private:
	typedef enum {
		UNKNOWN, UPPER, LOWER, CENTER
	} regions;

	bool checkUpperLimit();
	bool checkLowerLimit();

	void processHoming();

	/*   local method to stop the motors  */
	void quickStop();

	/* local method that is called
	 * in process & setPower; it ensures
	 * that the shuttle is not set to 
	 * drive beyond a limit */
	void setSafeSpeed();

	//WeejusSingleOutput* arms;
	WeejusCANJaguar* arms;
	WeejusPotEncoder* pot_enc;

	// I wish for an "intimate:" section that hides
	// some things from itself

	// If there is a limit hit detection problem (things move too slowly), 
	// replace with a WeejusCounter and an AnalogTrigger class
	DigitalInput* top_limit;
	DigitalInput* bottom_limit;
	DigitalInput* middle_limit;

	float output_power;

	bool limits_broken;

	// if the motors may run
	bool enabled;

	// the region of the track in which the
	//  shuttle begins
	regions shuttle_region;
	// moving to the middle; location is known
	bool homing;

	bool hit_top_limit;
	bool hit_bottom_limit;
	bool under_center_limit;

};

#endif /*WEEJUS_CLIMBERDRIVE_H_*/
