#include "WeejusControls.h"
#include "DSMap.h"

#define SMART_DASH 1
#define USE_RING_FOR_FEED_LIGHT 1
#define USE_DEPLOY_MODE 1

WeejusControls* WeejusControls::_theControls = 0;

WeejusControls* WeejusControls::getInstance() {
	if (_theControls == NULL)
		_theControls = new WeejusControls();

	return _theControls;
}
WeejusControls::WeejusControls() {
	_lStick = new Joystick(DS_JOY_PORT_LEFT);
	_rStick = new Joystick(DS_JOY_PORT_RIGHT);
	_auxStick = new Joystick(DS_JOY_PORT_AUX);
	_extendedIO = new Joystick(DS_JOY_PORT_EMULATED);

	_driverStationLCD = DriverStationLCD::GetInstance();

	_drive = WeejusDrive::getInstance();
	_shooter = WeejusShooter::getInstance();
	_climber = WeejusClimber::getInstance();
	_target = WeejusTarget::getInstance();
	// do lights after we initialize the SmartDashboard  _lights = WeejusLights::getInstance();

	_buttonShoot = false;
	_buttonDeploy = false;
	_buttonFeed = false;
	_buttonReverse = false;
	_buttonLightRing = false;

	_deployedMode = false; // not deployed yet
	_deployedArmsOut = false; // arms not out yet

	_numDiscs = -1; // force it to update the first time
	_shooterAngleActual = -9000; // force it to update the first time
	_shooterAngleDesired = -9000; // force it to update the first time
	_shooterPower = -9000; // force it to update the first time
	_shooterSensorDebug = false;
	_startDiscs = 0;

	_lcdCleared = false;

	_storageBroken = false;
	_shooterBroken = false;
	_controlBroken = false;
	_liftBroken = false;
	_coastEnabled = false;
	_limitsBroken = false;
	_climbBroken = false;
	_shooterOverride = false;
#if SMART_DASH
	SmartDashboard::init();
	//initialize all SmartDashboard keys and values
	SmartDashboard::PutNumber(SMART_SHOOTER_POWER, 0.0f);
	SmartDashboard::PutNumber(SMART_ANGLE_ACTUAL, 0.0f);
	SmartDashboard::PutNumber(SMART_ANGLE_DESIRED, 0.0f);
	SmartDashboard::PutNumber(SMART_NUM_DISCS, 0);
	SmartDashboard::PutNumber(SMART_AUTO_DELAY, 0.0f);
	SmartDashboard::PutNumber(SMART_AUTO_MODE, 0);
	SmartDashboard::PutNumber(SMART_ACTUAL_DISCS, 0);

	_lights = new WeejusLights(true, USE_RING_FOR_FEED_LIGHT);
#else
	_lights = new WeejusLights(false, USE_RING_FOR_FEED_LIGHT);
#endif
}

WeejusControls::~WeejusControls() {
	delete _rStick;
	delete _lStick;
	delete _auxStick;
	delete _extendedIO;
	delete _lights;
}

void WeejusControls::forceEnableLights(bool on) {
	_lights->forceOn(on);
}

void WeejusControls::process(ROBOT_PROCESSES robotProcess) {
	int numDiscs;

	if (robotProcess == DISABLED) {
		// reset "flips" if we were in deployed mode
		if (_deployedMode) {
			_lights->forwardToggle(false);
			_target->flip(false);
		}
		_deployedMode = false; // not deployed yet
		_deployedArmsOut = false; // arms not out yet

		processBrokenSwitches();
		processExtendedIOStick(false);

		// number of discs to shooter. we only do this
		// here so that we only get the number 'once'
		// just before Autonomous or Tele-op loops run.
		// this works because the robot always starts
		// in Disabled mode before anything else.
#if SMART_DASH
		_startDiscs = (int) SmartDashboard::GetNumber(SMART_NUM_DISCS);
#else
		_startDiscs = 0;
#endif
		_shooter->setNumDiscs(_startDiscs);
	} else if (robotProcess == AUTONOMOUS) {
#if SMART_DASH
		if (_startDiscs != _numDiscs) {
			_startDiscs = _numDiscs;
			SmartDashboard::PutNumber(SMART_NUM_DISCS, _numDiscs);
		}
#endif
	} else if (robotProcess == TELEOP) {
#if SMART_DASH
		if (_startDiscs != _numDiscs) {
			_startDiscs = _numDiscs;
			SmartDashboard::PutNumber(SMART_NUM_DISCS, _numDiscs);
		}
#endif
		processBrokenSwitches();
		processAuxButtons();
		processJoySticks();
		processAuxStick();
		processExtendedIOStick(true);
	}

#if SMART_DASH
	// only update dashboard it if changed
	numDiscs = _shooter->getNumDiscs();
	if (numDiscs != _numDiscs) {
		_numDiscs = numDiscs;
		SmartDashboard::PutNumber(SMART_ACTUAL_DISCS, _numDiscs);
	}
#endif

	_lights->process();
	processDebug();
}

void WeejusControls::getAutoValues(int *mode, int *delay) {
	*mode = 0;
	*delay = 0;

	// if lift is broken, then we cannot shoot and since
	// all autonomous modes shoot, we will catch that in
	// here versus WeejusAuto.cpp
	if (_liftBroken) {
		printf("[Controls] lift broken, so no auto-mode\n");
		return;
	}

#if SMART_DASH
	*mode = (int) SmartDashboard::GetNumber(SMART_AUTO_MODE);
	*delay = (int) SmartDashboard::GetNumber(SMART_AUTO_DELAY);
#endif
}

void WeejusControls::processAuxButtons() {
	bool buttonPressed;

	// only do shooter stuff if not climbing
	if (!_deployedMode) {
		// check deploy button
		buttonPressed = (_auxStick->GetRawButton(DS_DEPLOY) == 1);
		if (_buttonDeploy != buttonPressed) {
			// if just pressed, then deploy the arms
			if (buttonPressed) {
#if USE_DEPLOY_MODE
				_deployedMode = true;
				_deployedArmsOut = false;
				_shooter->stopShoot(); // stop incase we are shooting
				_shooter->moveToAngle(ANGLE_CLIMB_TARGET);
				_lights->forwardToggle(true);
				_target->flip(true);
				// deployArms() is called below (after this !_deployedMode if statement
#else
				if (!_climber->deployArms())
				{
					// unable to deploy, so flash the lights
					_lights->lightRave(WeejusLights::raveDEPLOY_FAIL);
				}
#endif
			}
		}
		_buttonDeploy = buttonPressed;

		// handle shooting button
		buttonPressed = (_auxStick->GetRawButton(DS_SHOOT_DISK) == 1);
		if (_buttonShoot != buttonPressed) {
			if (buttonPressed) {
				_shooter->startShoot();
			} else {
				_shooter->stopShoot();
			}
		}
		_buttonShoot = buttonPressed;

		// manually move fan blade, must always call climber each loop
		buttonPressed = (_auxStick->GetRawButton(DS_INDEX_DISK) == 1);
		if (buttonPressed && _storageBroken) {
			_shooter->turnIndexer();
		}

		// switch feed mode
		buttonPressed = (_auxStick->GetRawButton(DS_FEED_MODE) == 1);
		if (_buttonFeed != buttonPressed) {
			// controls ignores Feed button if lift broken
			if (buttonPressed && !_liftBroken) {
				_shooter->moveToAngle(ANGLE_FEED);
				// NOTE: do not send to dashboard
			}
			_lights->feedToggle(buttonPressed);
			_shooter->feedMode(buttonPressed);
		}
		_buttonFeed = buttonPressed;
	}
#if USE_DEPLOY_MODE
	// if we deployed but arms not out, then try again. we need
	// to re-try if shooter isn't all the way up when Deploy
	// button is pressed, so since _climber will wait for
	// the arms to be up, we have keep re-trying
	if (_deployedMode && !_deployedArmsOut) {
		// climber hasn't deployed arms yet (probably because
		// the shooter isn't up yet)
		if (!_climber->deployArms()) {
			// unable to deploy, so flash the lights
			_lights->lightRave(WeejusLights::raveDEPLOY_FAIL);
		} else {
			_deployedArmsOut = true;
		}

	}
#endif

	// handle camera tilt buttons
	if (_auxStick->GetRawButton(DS_TILT_CAMERA_DOWN) == 1) {
		_target->tiltDown(CAMERA_ANGLE_AMOUNT);
	}
	if (_auxStick->GetRawButton(DS_TILT_CAMERA_UP) == 1) {
		_target->tiltUp(CAMERA_ANGLE_AMOUNT);
	}

	//controls light ring
	buttonPressed = (_auxStick->GetRawButton(DS_CONTROL_LIGHT_RING) == 1);
#if !USE_RING_FOR_FEED_LIGHT
	if (_buttonLightRing != buttonPressed)
	{
		_target->setLight(buttonPressed);
	}
#endif
	_buttonLightRing = buttonPressed;
}
void WeejusControls::processJoySticks() {
	float lpower = getPower(_lStick);
	float rpower = getPower(_rStick);
	bool buttonPressed;

	// handle reverse buttons (only flip if not already deployed)
	buttonPressed = _lStick->GetRawButton(DS_JOYBTN_REVERSE)
			|| _rStick->GetRawButton(DS_JOYBTN_REVERSE);
	if ((_buttonReverse != buttonPressed) && !_deployedMode) {
		// only update lights and camera when button changes state
		_lights->forwardToggle(buttonPressed);
		_target->flip(buttonPressed);
	}
	_buttonReverse = buttonPressed;
	if (_buttonReverse) {
		lpower = -lpower;
		rpower = -rpower;
		_drive->setMotorSpeed(rpower, lpower);
	} else {
		_drive->setMotorSpeed(lpower, rpower);
	}

	// fun buttons
	if (_rStick->GetRawButton(DS_JOYBTN_FUN_1)) {
		_lights->lightRave(WeejusLights::raveFUN1);
	}
	if (_rStick->GetRawButton(DS_JOYBTN_FUN_2)) {
		_lights->lightRave(WeejusLights::raveFUN2);
	}

	// dumper deploy;
	if (_deployedMode) {
		if (_rStick->GetRawButton(DS_DUMPER_DEPLOY) || _auxStick->GetRawButton(
				DS_DUMPER_DEPLOY)) {
			_climber->dump();
		}
	}
}

void WeejusControls::processAuxStick() {
	// allow aux driver to control arms if auto-climb is broken 
	float power;
	power = _auxStick->GetY();
	//Deadzone, prevents motors from running
	if (power > -.05 && power < .05)
		power = 0.0f;

	// we send positive for shuttle up;
	// this is always called, even when broken;
	//   the _climber deals with it on its
	//   own side
	_climber->moveArms(-power,
			_auxStick->GetRawButton(DS_CLIMB_PREDEP_OVERRIDE));
}

void WeejusControls::processDebug() {
	bool buttonPressed;

	// print shooter flag sensor debug 
	// this is we just need to see the toggle
	buttonPressed = (_lStick->GetRawButton(DS_JOYBTN_DEBUG_FLAGS) == 1);
	if (_shooterSensorDebug != buttonPressed) {
		_shooter->setSensorDebug(buttonPressed);
	}
	_shooterSensorDebug = buttonPressed;

	if (_lStick->GetRawButton(DS_JOYBTN_DEBUG_SHOOTER)) {
		_driverStationLCD->Clear();
		_driverStationLCD->PrintfLine(DriverStationLCD::kUser_Line1,
				_shooter->getDebugLine(1));
		_driverStationLCD->PrintfLine(DriverStationLCD::kUser_Line2,
				_shooter->getDebugLine(2));
		_driverStationLCD->PrintfLine(DriverStationLCD::kUser_Line3,
				_shooter->getDebugLine(3));
		_driverStationLCD->PrintfLine(DriverStationLCD::kUser_Line4,
				_shooter->getDebugLine(4));
		_driverStationLCD->PrintfLine(DriverStationLCD::kUser_Line5,
				_shooter->getDebugLine(5));
		_driverStationLCD->PrintfLine(DriverStationLCD::kUser_Line6,
				_shooter->getDebugLine(6));
		_driverStationLCD->UpdateLCD();
		_lcdCleared = false;
	} else if (_lStick->GetRawButton(DS_JOYBTN_DEBUG_CLIMBER)) {
		_driverStationLCD->Clear();
		_driverStationLCD->PrintfLine(DriverStationLCD::kUser_Line1,
				_climber->getDebugLine(1));
		_driverStationLCD->PrintfLine(DriverStationLCD::kUser_Line2,
				_climber->getDebugLine(2));
		_driverStationLCD->PrintfLine(DriverStationLCD::kUser_Line3,
				_climber->getDebugLine(3));
		_driverStationLCD->PrintfLine(DriverStationLCD::kUser_Line4,
				_climber->getDebugLine(4));
		_driverStationLCD->PrintfLine(DriverStationLCD::kUser_Line5,
				_climber->getDebugLine(5));
		_driverStationLCD->PrintfLine(DriverStationLCD::kUser_Line6,
				_climber->getDebugLine(6));
		_driverStationLCD->UpdateLCD();
		_lcdCleared = false;
	} else if (_lStick->GetRawButton(DS_JOYBTN_DEBUG_TARGET)) {
		_driverStationLCD->Clear();
		_driverStationLCD->PrintfLine(DriverStationLCD::kUser_Line1,
				_target->getDebugLine(1));
		_driverStationLCD->PrintfLine(DriverStationLCD::kUser_Line2,
				_target->getDebugLine(2));
		_driverStationLCD->PrintfLine(DriverStationLCD::kUser_Line3,
				_target->getDebugLine(3));
		_driverStationLCD->PrintfLine(DriverStationLCD::kUser_Line4,
				_target->getDebugLine(4));
		_driverStationLCD->PrintfLine(DriverStationLCD::kUser_Line5,
				_target->getDebugLine(5));
		_driverStationLCD->PrintfLine(DriverStationLCD::kUser_Line6,
				_target->getDebugLine(6));
		_driverStationLCD->UpdateLCD();
		_lcdCleared = false;
	} else if (_lStick->GetRawButton(DS_JOYBTN_DEBUG_DRIVE)) {
		_driverStationLCD->Clear();
		_driverStationLCD->PrintfLine(DriverStationLCD::kUser_Line1,
				_drive->getDebugLine(1));
		_driverStationLCD->PrintfLine(DriverStationLCD::kUser_Line2,
				_drive->getDebugLine(2));
		_driverStationLCD->PrintfLine(DriverStationLCD::kUser_Line3,
				_drive->getDebugLine(3));
		_driverStationLCD->PrintfLine(DriverStationLCD::kUser_Line4,
				_drive->getDebugLine(4));
		_driverStationLCD->PrintfLine(DriverStationLCD::kUser_Line5,
				_drive->getDebugLine(5));
		_driverStationLCD->PrintfLine(DriverStationLCD::kUser_Line6,
				_drive->getDebugLine(6));
		_driverStationLCD->UpdateLCD();
		_lcdCleared = false;
	} else if (!_lcdCleared) {
		_driverStationLCD->Clear();
		_driverStationLCD->UpdateLCD();
		_lcdCleared = true;
	}
}

void WeejusControls::processExtendedIOStick(bool teleop) {
	float slider;
	int sliderInt;

	// process the angle slider
	slider = -_extendedIO->GetY();
	if (!_buttonFeed && teleop && (!_deployedMode || _liftBroken)) // really move only in tele-op & not climbing
	{
		_shooter->moveToAngle(slider);
	}
#if SMART_DASH
	// always send to dashboard if it has changed. and if it
	// has changed enough that the user will even
	// notice (i.e. only 2 decimal places)
	sliderInt = convertSlider(slider);
	if (sliderInt != _shooterAngleDesired) {
		_shooterAngleDesired = sliderInt;
		SmartDashboard::PutNumber(SMART_ANGLE_DESIRED,
				(double) _shooterAngleDesired);
	}
#endif

#if SMART_DASH
	// only update angle if it has changed. and if it
	// has changed enough that the user will even
	// notice (i.e. only 2 decimal places)
	slider = _shooter->calcShooterAngle();
	sliderInt = convertSlider(slider);
	if (sliderInt != _shooterAngleActual) {
		_shooterAngleActual = sliderInt;
		SmartDashboard::PutNumber(SMART_ANGLE_ACTUAL,
				(double) _shooterAngleActual);
	}
#endif

	// process power slider
	slider = _shooter->linearRangeScale(_extendedIO->GetX(), -1.0f, 1.0f, 0.3f,
			1.0f);
	//slider = ((slider - SHOOTER_SLIDER_MIN)/(SHOOTER_SLIDER_MAX - SHOOTER_SLIDER_MIN));
	if (teleop && !_deployedMode) // really set speed only in tele-op & not climbing
	{
		_shooter->setSpeed(slider); // convert to 0.0 to 1.0 for shooter
	}
#if SMART_DASH
	// 0 to 100 for dashboard, but dashboard will convert it
	sliderInt = convertSlider(slider);
	if (sliderInt != _shooterPower) {
		_shooterPower = sliderInt;
		SmartDashboard::PutNumber(SMART_SHOOTER_POWER, (double) _shooterPower);
	}
#endif
}

// convert slider value to int between 0 to 100
int WeejusControls::convertSlider(float slider) {
	int sliderInt;

	// round to nearest 2 decimal places
	if (slider < 0) {
		slider -= 0.005f;
	} else if (slider > 0) {
		slider += 0.005f;
	}

	sliderInt = (int) (slider * 100.0f);
	if (sliderInt < -100) {
		sliderInt = -100;
	} else if (sliderInt > 100) {
		sliderInt = 100;
	}
	return (sliderInt);
}

void WeejusControls::processBrokenSwitches() {
	bool buttonPressed;

	// manually move fan blade
	buttonPressed = (_extendedIO->GetRawButton(DS_BROKEN_STORAGE) == 1);
	if (_storageBroken != buttonPressed) {
		_shooter->setStorageBroken(buttonPressed);
		_lights->setStorageBroken(buttonPressed);
	}
	_storageBroken = buttonPressed;

	// allows for manual power input
	buttonPressed = (_extendedIO->GetRawButton(DS_BROKEN_SHOOTER) == 1);
	if (_shooterBroken != buttonPressed) {
		_shooter->setEncoderBroken(buttonPressed);
	}
	_shooterBroken = buttonPressed;

	// everything has died, no shooting, shoot, angle, or motor functions
	buttonPressed = (_extendedIO->GetRawButton(DS_BROKEN_SHOOTER_STORAGE) == 1);
	if (_controlBroken != buttonPressed) {
		_shooter->setDead(buttonPressed);
	}
	_controlBroken = buttonPressed;

	// manually regulate angle
	buttonPressed = (_extendedIO->GetRawButton(DS_BROKEN_LIFT) == 1);
	if (_liftBroken != buttonPressed) {
		_shooter->setAngleBroken(buttonPressed);
	}
	_liftBroken = buttonPressed;

	// manually align and climb 
	buttonPressed = (_extendedIO->GetRawButton(DS_BROKEN_AUTO_CLIMB) == 1);
	if (_coastEnabled != buttonPressed) {
		_drive->setCoast(buttonPressed);
	}
	_coastEnabled = buttonPressed;

	// set the climber drive pot broken
	buttonPressed = (_extendedIO->GetRawButton(DS_BROKEN_ARM_LIMITS) == 1);
	if (_limitsBroken != buttonPressed) {
		_climber->setLimitsBroken(buttonPressed);
	}
	_limitsBroken = buttonPressed;

	// no climbing allowed ever until stuff is fixed
	buttonPressed
			= (_extendedIO->GetRawButton(DS_BROKEN_CLIMB_COMPLETELY) == 1);
	if (_climbBroken != buttonPressed) {
		_climber->setDead(buttonPressed);
	}
	_climbBroken = buttonPressed;

	//keep shooter wheel running when override is toggled
	buttonPressed = (_extendedIO->GetRawButton(DS_BROKEN_SPEED_OVERRIDE) == 1);
	if (_shooterOverride != buttonPressed) {
		_shooter->shootOverride(buttonPressed);
	}
	_shooterOverride = buttonPressed;
}

float WeejusControls::getPower(Joystick * stick) {
	float power = stick->GetY();

	if (power > -.05 && power < .05) //Deadzone, prevents motors from running
		power = 0; //when not touching the joysticks

	power = power * power * power; //Cube the power to ramp input

	if (stick->GetRawButton(DS_JOYBTN_SLOW))
		power *= .3; //Cut power to 3/10 for slower driving
	else if (stick->GetRawButton(DS_JOYBTN_TURBO)) {
	}
	//Do nothing, keep full power
	else
		power *= .7; //Cut power to 7/10 for normal operation

	return power;
}
