#ifndef _CONTROLS_H_
#define _CONTROLS_H_
#include "WPILib.h"
#include "WeejusDrive.h"
#include "WeejusShooter.h"
#include "WeejusClimber.h"
#include "WeejusTarget.h"
#include "WeejusLights.h"

/*Controls Class
 - get angle from slider and send to shooter
 - get suggested angle from "Targeting" and display on SmartDashboard
 - tilt camera from driver, goes to "Targeting"
 - get "reverse" camera and do "Targeting"
 -auto-targeting goes to "Targeting", and Controls stops call Shooter [FUTURE]
 */

// the extent of the shooter in degress 
#define ANGLE_DEGREES_MIN 		10.0f
#define ANGLE_DEGREES_MAX 		80.0f

// the angle slider min and max values
#define ANGLE_SLIDER_MIN  		-1.0f
#define ANGLE_SLIDER_MAX  		1.0f

// the part of the angle slider that is the dead zone for broken mode
#define ANGLE_SLIDER_DEAD_MIN	-0.2f
#define ANGLE_SLIDER_DEAD_MAX	0.2f

// the shooter slider min and max values
#define SHOOTER_SLIDER_MIN  	-1.0f
#define SHOOTER_SLIDER_MAX  	1.0f

// camera degree amount for each process loop
#define CAMERA_ANGLE_AMOUNT		0.75f

class WeejusControls {
public:
	enum ROBOT_PROCESSES {
		DISABLED, AUTONOMOUS, TELEOP
	};

	/*
	 * Method: getInstance();
	 * 
	 * Inputs: Null
	 * Outputs: pointer to the WeejusControls object
	 * 
	 * Remarks:
	 * returns the one and only WeejusControls
	 * if it does not exist, it will be created
	 * otherwise it will just return the existing instance
	 */
	static WeejusControls* getInstance();

	~WeejusControls();

	void forceEnableLights(bool on);

	/*
	 * Method: process()
	 * 
	 * Inputs: ROBOT_PROCESSES - which process state the robot is in
	 * Outputs: Null
	 * 
	 * Remarks:
	 * updates all internal status values and performs all actions
	 * MUST be called at the end of each loop
	 * 
	 * 
	 */
	void process(ROBOT_PROCESSES robotProcess);

	/*
	 * Method: getAutoValues()
	 * 
	 * Inputs:
	 * 		mode		gets the auto-mode from driver station
	 * 		delay		gets the delay from the driver station
	 * Outputs: Null
	 * 
	 */
	void getAutoValues(int *mode, int *delay);

private:
	WeejusControls();
	void processBrokenSwitches();
	void processAuxButtons();
	void processJoySticks();
	void processAuxStick();
	void processExtendedIOStick(bool teleop);
	void processDebug();
	float getPower(Joystick* stick);

	int convertSlider(float slider);

	static WeejusControls* _theControls;

	Joystick *_rStick;
	Joystick *_lStick;
	Joystick *_auxStick;
	Joystick *_extendedIO;

	DriverStationLCD* _driverStationLCD;

	WeejusDrive *_drive;
	WeejusShooter *_shooter;
	WeejusClimber *_climber;
	WeejusTarget *_target;
	WeejusLights *_lights;

	bool _buttonShoot;
	bool _buttonDeploy;
	bool _buttonFeed;
	bool _buttonReverse;
	bool _buttonLightRing;

	bool _deployedMode; // deploy button pressed, so no longer shooting
	bool _deployedArmsOut; // arms have been released (if not keep trying in _deployedMode

	int _numDiscs;
	int _shooterAngleDesired;
	int _shooterAngleActual;
	int _shooterPower;
	int _startDiscs;

	bool _shooterSensorDebug;

	bool _lcdCleared;

	bool _storageBroken;
	bool _shooterBroken;
	bool _controlBroken;
	bool _liftBroken;
	bool _coastEnabled;
	bool _limitsBroken;
	bool _climbBroken;
	bool _shooterOverride;
};
#endif
