#include "WeejusDrive.h"
#include <math.h>

// Controller factors for driving (movement)
// Want output to be 1.0 (max power) when distance >= 20 inches, so P = 1/20  
#define DRIVE_P  (1.0/20.0)
// Minimum power to motors during auto drive
#define DRIVE_MIN 0.25
// auto drive acceleration rate - motor power per second
#define DRIVE_ACCEL 1.0
// Multiplicative factor applied to angular error when driving straight
// Amount of motor power boost per degree of error
#define KEEP_STRAIGHT_CORRECTION_FACTOR 0.05

// amount to scale encoder values from the left encoder
// to be in sync with the right encoder.
// needed when the encoders are different types.
// if both encoders are the same, set this to 1.0f
#define LEFT_ENCODER_SCALE	1.5f

// Controller factors for turning
// Want output to be 1.0 (max turn) when error >= 45 degrees, so P = 1/45 
#define TURN_P  (1.0/45.0)
#define TURN_MIN 0.25

// Alignment numbers
#define ALIGN_TURN_MIN_SPEED (0.1)      // real drive power on the slower side
#define ALIGN_TURN_MAX_SPEED (0.5)      // real drive power on the faster side
#define ALIGN_STRAIGHT_SPEED (0.3)      // real drive power going straight
#define ALIGN_CREEP_SPEED    (0.05)     // creeping speed, to counteract the push-away
// effect when close to the pyramid

WeejusDrive* WeejusDrive::theDrivep = 0;

WeejusDrive::WeejusDrive() :
	lencoder(DIG_DRIVETRAIN_LEFT_ENCODER_A),
			rencoder(DIG_DRIVETRAIN_RIGHT_ENCODER_A),

			leftMotor(CANID_DRIVETRAIN_LEFT),
			rightMotor(CANID_DRIVETRAIN_RIGHT),

			autoState(AUTO_NONE), gyro(ANA_DRIVETRAIN_GYRO) {
	bool motorSafety = false;
	leftMotor.SetSafetyEnabled(motorSafety);
	rightMotor.SetSafetyEnabled(motorSafety);

	// Set default motor neutral mode
	neutralMode = WeejusCANJaguar::kNeutralMode_Brake;
	rightMotor.ConfigNeutralMode(neutralMode);
	leftMotor.ConfigNeutralMode(neutralMode);

	// Zero the sensors
	lencoder.Reset();
	lencoder.SetReverseDirection(true);
	lencoder.Start();
	rencoder.Reset();
	rencoder.SetReverseDirection(false);
	rencoder.Start();

	// TODO confim gyro sensitivy for 2013 - which PDF?
	gyro.SetSensitivity(0.007f);
	gyro.Reset();

	autoTimer = new Timer();

	lastLeftPower = 0.0;
	lastRightPower = 0.0;

}

WeejusDrive* WeejusDrive::getInstance() {
	if (theDrivep == 0)
		theDrivep = new WeejusDrive();
	return theDrivep;
}

//apparently this should never be called
WeejusDrive::~WeejusDrive() {
	delete autoTimer;
}

void WeejusDrive::initialize() {
	lastLeftPower = 0.0f;
	lastRightPower = 0.0f;
	autoState = AUTO_NONE;

#if USING_OUTPUT_THROTTLING
	powerThrottleAmount = 0.0f;
#endif

	autoTimer->Reset();
	gyro.Reset();
	rencoder.Reset();
	lencoder.Reset();

	leftMotor.Set(0.0f);
	rightMotor.Set(0.0f);
}

void WeejusDrive::process() {
	switch (autoState) {
	case AUTO_DRIVE: {
		int distLeft = autoTargetDistance - getEncoderAvg();
		//printf("processDrive - distLeft:%i, getEncoderAvg:%i\n", distLeft, getEncoderAvg());
		if (distLeft <= DISTANCE_CLOSE_ENOUGH) {
			setMotorSpeedInternal(0, 0);
			autoState = AUTO_NONE;
			printf("processDrive done - distLeft:%i, getEncoderAvg:%i\n",
					distLeft, getEncoderAvg());
		} else {
			double pPow = distLeft * DRIVE_P;
			double accelPow = autoTimer->Get() * DRIVE_ACCEL;
			if (pPow > autoMaxPower)
				pPow = autoMaxPower;

			// pPow is forward-motion power. Now gyro correct
			// negative delta means robot pointed left-of target
			double delta = getGyro() - autoTargetAngle;
			delta *= KEEP_STRAIGHT_CORRECTION_FACTOR;
			// Swap delta direction -- negative in driveAndTurn means turn left
			delta *= -1.0f;

			driveWithCorrection(pPow, -delta, DRIVE_MIN, accelPow);
		}
		break;
	}
	case AUTO_TURN: {
		// negative delta means robot pointed left-of target
		double delta = getGyro() - autoTargetAngle;
		if (fabs(delta) < ANGLE_CLOSE_ENOUGH) {
			setMotorSpeedInternal(0, 0);
			autoState = AUTO_NONE;
		} else {
			double turnPow = delta * TURN_P;
			if (turnPow < 0 && turnPow > -TURN_MIN)
				turnPow = -TURN_MIN;
			if (turnPow > 0 && turnPow < TURN_MIN)
				turnPow = TURN_MIN;
			// Negative turnpow means turn to right ->
			//     so push left positive and right negative
			// Positive turnpow means turn left
			//     so push left negative and right positive
			// Therefore, invert left
			setMotorSpeedInternal(-turnPow, turnPow);
		}
		break;
	}
	case AUTO_DRIVE_TURN:
		// Not yet implemented
		autoState = AUTO_NONE;
		break;
	default:
		// Do nothing - manual is handled directly
		break;
	}

#if USING_OUTPUT_THROTTLING
	float lc = leftMotor.GetOutputCurrent();
	float rc = rightMotor.GetOutputCurrent();
	if ((lc > OT_LIMIT_THRESHOLD) || (rc > OT_LIMIT_THRESHOLD)) {
#if DEBUG_OUTPUT_THROTTLING
		printf("[Drive] Power throttle is at %f\n", powerThrottleAmount);
#endif
		powerThrottleAmount += OT_CUTTING_RATE;
	} else if (powerThrottleAmount > 0.0f) {
		if ((lc < OT_CLEAR_THRESHOLD) && (rc < OT_CLEAR_THRESHOLD)) {
#if DEBUG_OUTPUT_THROTTLING
			printf("[Drive] Power throttle reset to zero\n");
#endif
			powerThrottleAmount = 0.0f;
		}
	}
#endif
}

void WeejusDrive::setMotorSpeed(float leftPower, float rightPower) {
	setMotorSpeedInternal(leftPower, rightPower);
	autoState = AUTO_NONE;
}

void WeejusDrive::setMotorSpeedInternal(float leftPower, float rightPower) {
	// Make sure power is in the range 1.0 to -1.0
	if (rightPower > 1.0)
		rightPower = 1.0;
	else if (rightPower < -1.0)
		rightPower = -1.0;
	if (leftPower > 1.0)
		leftPower = 1.0;
	else if (leftPower < -1.0)
		leftPower = -1.0;
#if USING_OUTPUT_THROTTLING
#if DEBUG_OUTPUT_THROTTLING
	printf("[Drive] Power: L:%f, R:%f   ->   ", leftPower, rightPower);
#endif
	if (leftPower > powerThrottleAmount) {
		// decrease positive power, but do not go below zero
		leftPower -= powerThrottleAmount;
	} else if (leftPower < -powerThrottleAmount) {
		// decrease negative power, but do not go above zero
		leftPower += powerThrottleAmount;
	}
	if (rightPower > powerThrottleAmount) {
		// decrease positive power, but do not go below zero
		rightPower -= powerThrottleAmount;
	} else if (rightPower < -powerThrottleAmount) {
		// decrease negative power, but do not go above zero
		rightPower += powerThrottleAmount;
	}
#if DEBUG_OUTPUT_THROTTLING
	printf("L:%f, R:%f\n", leftPower, rightPower);
#endif
#endif

	rightMotor.Set(-rightPower);
	leftMotor.Set(leftPower);
	lastLeftPower = leftPower;
	lastRightPower = rightPower;
}

static int inchesToTicks(float inches) {
	return (int) (inches * ENCODER_COUNTS_PER_INCH);
}

void WeejusDrive::autonomousMove(float distance, float maxPower) {
	printf("autonomousMove - dist:%f, maxSpeed:%f\n", distance, maxPower);

	// Cancel any move or turn actions in progress
	stopAction();

	autoState = AUTO_DRIVE;
	autoMaxPower = maxPower;
	if (distance > 0) {
		autoDirection = -1.0f;
	} else {
		distance = -distance;
		autoDirection = 1.0f;
	}
	autoTargetDistance = getEncoderAvg() + inchesToTicks(distance);
	autoTargetAngle = getGyro();
	printf("autonomousMove - getEncoderAvg:%i targetdist:%i, targetangle:%f\n",
			getEncoderAvg(), autoTargetDistance, autoTargetAngle);
	autoTimer->Reset();
	autoTimer->Start();

	return;

}

void WeejusDrive::autonomousTurn(float degrees) {
	stopAction(); // Cancel any actions in progress	

	autoState = AUTO_TURN;
	autoMaxPower = 1.0;
	autoTargetDistance = getEncoderAvg() + 0;
	autoTargetAngle = getGyro() + degrees;
	autoTimer->Reset();
	autoTimer->Start();

	return;
}

void WeejusDrive::pyramidAlignmentMotion(AlignMotion motion) {
	switch (motion) {
	default:
		break;
	case Backward:
		setMotorSpeedInternal(ALIGN_STRAIGHT_SPEED, ALIGN_STRAIGHT_SPEED);
		break;
	case RightBackTurn:
		setMotorSpeedInternal(ALIGN_TURN_MAX_SPEED, ALIGN_TURN_MIN_SPEED);
		break;
	case LeftBackTurn:
		setMotorSpeedInternal(ALIGN_TURN_MIN_SPEED, ALIGN_TURN_MAX_SPEED);
		break;
	case BackwardCreep:
		setMotorSpeedInternal(ALIGN_CREEP_SPEED, ALIGN_CREEP_SPEED);
		break;

	}
}

bool WeejusDrive::doneYet() {
	return autoState == AUTO_NONE;
}

INT32 WeejusDrive::getEncoderAvg() {
	INT32 left;

	left = (int) (((double) lencoder.Get()) * LEFT_ENCODER_SCALE);
	return (rencoder.Get() + left) / 2;
}

float WeejusDrive::getGyro(void) {
	return gyro.GetAngle();
}

void WeejusDrive::setCoast(bool coast) {
	if (coast)
		neutralMode = WeejusCANJaguar::kNeutralMode_Coast;
	else
		neutralMode = WeejusCANJaguar::kNeutralMode_Brake;

	rightMotor.ConfigNeutralMode(neutralMode);
	leftMotor.ConfigNeutralMode(neutralMode);
	return;
}

void WeejusDrive::stopAction() {
	autoState = AUTO_NONE;
	setMotorSpeedInternal(0, 0);

	return;
}
char* WeejusDrive::AutoStateToString(WeejusDrive::AutoState state) {
	switch (state) {
	case WeejusDrive::AUTO_DRIVE:
		return "DRIVE";
	case WeejusDrive::AUTO_TURN:
		return "TURN";
	case WeejusDrive::AUTO_DRIVE_TURN:
		return "DRIVE TURN";
	case WeejusDrive::AUTO_NONE:
	default:
		return "NONE";
	}
}
char* WeejusDrive::getDebugLine(int line) {
	switch (line) {
	case 1:
		snprintf(_dbgBuf, sizeof(_dbgBuf), "ENC L: %i R: %i", lencoder.Get(),
				rencoder.Get());
		break;
	case 2:
		snprintf(_dbgBuf, sizeof(_dbgBuf), "LMOT: %1.4f C: %3.4f",
				(double) leftMotor.Get(), (double) leftMotor.GetOutputCurrent());
		break;
	case 3:
		snprintf(_dbgBuf, sizeof(_dbgBuf), "RMOT: %1.4f C: %3.4f",
				(double) rightMotor.Get(),
				(double) rightMotor.GetOutputCurrent());
		break;
	case 4:
		snprintf(_dbgBuf, sizeof(_dbgBuf), "GYRO: %f", (double) gyro.GetAngle());
		break;
	case 5:
		snprintf(_dbgBuf, sizeof(_dbgBuf), "STATE: %s",
				AutoStateToString(autoState));
		break;
	case 6:
	default:
		snprintf(_dbgBuf, sizeof(_dbgBuf), " ");
		break;

	}
	return _dbgBuf;
}

// driveAndTurn - Set motor speeds to drive and turn in autonomous mode
void WeejusDrive::driveWithCorrection(float power, float turn, float min,
		float max) {
	// IF turn == 0, left and right power should both equal the power value
	// IF power == 0 and turn == 0, left and right power should be zero
	// IF power == 0 and turn = -1, (full left turn in place) right power should be -1 and left power should be +1
	// IF power == 0 and turn = 1, (full right turn in place) left power should be -1 and right power should be +1
	// BUT IF power = 1 and turn = 1 (full speed right turn) left=0, right =1

	// Start with each equal to power
	float left = power;
	float right = power;

	// Add/subtract turn value
	left = left - turn;
	right = right + turn;

	// Clamp
	if (left > max)
		left = max;
	else if (left < min)
		left = min;
	if (right > max)
		right = max;
	else if (right < min)
		right = min;

	// Apply the computed power to the drive motors
	setMotorSpeedInternal(autoDirection * left, autoDirection * right);
}
