//arm the (include) guards
#ifndef _DRIVE_H_
#define _DRIVE_H_
#include "WPILib.h"
#include "WeejusCANJaguar.h"
#include "IOmap.h"
#include "DSmap.h"
#
// motors are run in current mode
#define USING_OUTPUT_THROTTLING 1
#define DEBUG_OUTPUT_THROTTLING 0

#define OT_LIMIT_THRESHOLD 40.0f// amps
#define OT_CLEAR_THRESHOLD 35.0f// amps
#define OT_CUTTING_RATE    0.01f// %vbus/code cycle
#define ENCODER_COUNTS_PER_INCH (42)					// TODO need to get this from measurement
// How close to setpoint to be considered done?
// In inches
#define DISTANCE_CLOSE_ENOUGH (0.5)
// In degrees
#define ANGLE_CLOSE_ENOUGH (1.0)
class WeejusDrive {
public:
	typedef enum {
		Backward, RightBackTurn, LeftBackTurn, BackwardCreep
	} AlignMotion;

	// getInstance - Static function returns a pointer to the one and only WeejusDrive 
	// object in the system.
	// Use it like this:
	//
	//      WeejusDrive* tdp; pointer to the system WeejusDrive object;
	//      tdp = WeejusDrive::getInstance();
	//      .  .  .
	//      tdp->setMotorSpeed(. . .);

	static WeejusDrive* getInstance();

	~WeejusDrive();

	//initialize - revert state variables to the 
	//             default at start of teleop
	void initialize();

	void process();

	//setMotorSpeed - Set motor speeds for teleoperated mode
	//        Inputs:
	//                leftPower: Desired power for left motor, range: -1.0 -> 1.0
	//                rightPower: Desired power for right motor, range: -1.0 -> 1.0
	//        Returns: nothin'
	void setMotorSpeed(float leftPower, float rightPower);

	//autonomousMove - begins a straight line movement a certain distance in inches
	//        Inputs:
	//                distance: the distance in inches that the robot must move
	//                        positive values: forwards
	//                        negative values: backwards
	//                maxPower: max power value for the motors
	//        Returns: Nothin'
	//        Note: If a previous autonomousMove() operation is under way, resets stopping point to "distance" from the present point
	//            If a previous autonomousTurn() operation is under way, cancels the turn, and then moves
	//            doneYet() function used to check completion
	void autonomousMove(float distance, float maxPower);

	//autonomousTurn - begins a turn in place a certain angle from the starting direction
	//        Inputs:
	//                degrees: The angle at which to turn, in degrees
	//                        positive value: turn to the right
	//                        negative value: turn to the left
	//        Returns: Nothing
	//        Note: The same applies to autonomousTurn as does autonomousMove. If a previous operation of autonomousTurn
	//            or autonomousMove is still in progress with autonomousTurn is called, it will interrupt the current
	//            action and start a new autonomousTurn from the location and angle it is currently at.
	void autonomousTurn(float degrees);

	//pyramidAlignmentMotion - sets the motor speeds accordingly make an alignment motion
	//        Inputs:
	//                motion: one of
	//                        Forward
	//                        RightTurn
	//                        LeftTurn
	//		  Returns: Nothing
	//        Note: This merely sets motor speeds.
	//
	//
	void pyramidAlignmentMotion(AlignMotion motion);

	//doneYet - tells whether last called autonomous command was completed
	//            Inputs: none
	//            Returns:
	//                    True: Last command was completed
	//                    False: Last command is in process of completion
	//            Note: This can be used to check the result of the autonomousMove() and autonomousTurn() commands
	bool doneYet();

	/* getGyro - retrieves the current gyro value.
	 * 			Inputs: bool rightMotor
	 * 				true: returns the encoder value for the right motor
	 * 				false: returns the encoder value for the left motor
	 * 					If no value is given, the function defaults to the right motor
	 * 			Returns: Heading in degrees, as a float. 					
	 *          Note: The gyro is reset ONCE when the Weejusdrive object is instantiated. 
	 */
	float getGyro(void);

	// setCoast - Coast or brake when motor drive is zero?
	//          Inputs: coast
	//                        ciast if true, brake if false 
	void setCoast(bool coast);

	/*void stopAction - signals to end current action
	 * 		Inputs:
	 * 				Nothin'
	 * 		Returns: Nothin' 								*/
	void stopAction();

	/*
	 * Method: getDebugLine(int);
	 * 
	 * Inputs:
	 * 			int line: integer representing the debug line to get (range 1->6)
	 * Outputs: char array containing the debug line to print
	 * 
	 * Remarks:
	 * will return the line specified by the number passed to the parameter
	 * only can return 6 debug lines
	 * if line < 1, will return line 1
	 * if line > 6, will return line 6
	 */
	char* getDebugLine(int line);
private:

	WeejusDrive();

	//Sensors
	Counter lencoder; // left wheel encoder
	Counter rencoder; // right wheel encoder

	WeejusCANJaguar::NeutralMode neutralMode; // What do motors do when power = 0?

	// driveWithCorrection - Set motor speeds to drive and turn in autonomous mode
	//        Inputs:
	//                power: Desired power for motors, range: 0.0 -> 1.0
	//                turn:  Desired turn, range: -1.0 -> 1.0
	//                          -1.0 means max counterclockwise turn
	//                           0.0 means no turn
	//                          +1.0 means max clockwise turn
	//                min:   Minimum value for each motor
	//                max:   Maximum value for each motor
	//        Returns: nothin'
	//        Notes: min and max are specified in absolute value terms.
	//               max always wins when min/max overlap
	//               combined power and turn get autoDirection applied to them
	//                  after clamping but before output
	void driveWithCorrection(float power, float turn, float min, float max);

	/* getEncoderAvg - retrieves the average of the left and right encoder value.
	 * 			Returns: The average number of ticks, as an INT32. 	
	 * 			Note: The encoder is reset ONCE when the WeejusDrive object is instantiated.
	 */
	INT32 getEncoderAvg();

	// Autonomous drive state & support
	float autoMaxPower;
	int autoTargetDistance; // Ticks
	float autoTargetAngle; // Degrees
	float autoDirection; // -1 rev or 1 fwd
	Timer *autoTimer;

	static WeejusDrive* theDrivep;

	void setMotorSpeedInternal(float leftPower, float rightPower);

	float lastLeftPower;
	float lastRightPower;

#if USING_OUTPUT_THROTTLING
	float powerThrottleAmount;
#endif

	WeejusCANJaguar leftMotor;
	WeejusCANJaguar rightMotor;
	enum AutoState {
		AUTO_NONE, AUTO_DRIVE, AUTO_TURN, AUTO_DRIVE_TURN
	} autoState;
	Gyro gyro;

	char _dbgBuf[DEBUG_LINE_LENGTH];

	char* AutoStateToString(WeejusDrive::AutoState state);
};
#endif
