#include "WeejusLights.h"

// the list of I2C bank addresses for the lights
static UINT8 _i2cBankAddress[I2C_NUMBER_BANKS] = I2C_BANK_ADDRESS_INIT;

WeejusLights::WeejusLights(bool useSmartDashboard, bool useRingForFeedLight) {
	int i;

	// get pointer to other classes within out robot
	_shooter = WeejusShooter::getInstance();
	_climber = WeejusClimber::getInstance();
	_useSmartDashboard = useSmartDashboard;
	_useRingForFeedLight = useRingForFeedLight;

	// get the I2C information setup. for each
	// I2C bank, set the direction to be output
	// for all GPIO bits
	_digitalModule = DigitalModule::GetInstance(1);
	for (i = 0; i < I2C_NUMBER_BANKS; i++) {
#if USE_I2C
		_i2c[i] = _digitalModule->GetI2C(_i2cBankAddress[i]);
#else
		_i2c[i] = NULL;
#endif
		if ((_i2c[i] == NULL) || _i2c[i]->Write((UINT8) 0x00, (UINT8) 0x00)) {
			printf("ERROR setting I2C direction for %i\n",
					(int) _i2cBankAddress[i]);
		}
	}

	_forceAllOn = false;

	// if using Camera ring light for Feed light, initial to true
	// so it forces process() to turn off the light
	_ringFeedLight = true;
	_targetLight = true;

	// initialize all SmartDashboard ligths to OFF.
	// then is actually using the SmartDashboard, we have to make
	// sure the values are OFF; we do not do this, the SmartDashboard
	// might crash
	_shooterDirectionLightSmartDashboard = false;
	_climberDirectionLightSmartDashboard = false;
	_feedGreenLightSmartDashboard = false;
	_feedRedLightSmartDashboard = false;
	_disc1GreenLightSmartDashboard = false;
	_disc1RedLightSmartDashboard = false;
	_disc2GreenLightSmartDashboard = false;
	_disc2RedLightSmartDashboard = false;
	_disc3GreenLightSmartDashboard = false;
	_disc3RedLightSmartDashboard = false;
	_disc4GreenLightSmartDashboard = false;
	_disc4RedLightSmartDashboard = false;
	if (_useSmartDashboard) {
		SmartDashboard::PutNumber(SMART_SHOOTER_DIR_LIGHT, 0.0f);
		SmartDashboard::PutNumber(SMART_CLIMBER_DIR_LIGHT, 0.0f);
		SmartDashboard::PutNumber(SMART_FEED_GREEN_LIGHT, 0.0f);
		SmartDashboard::PutNumber(SMART_FEED_RED_LIGHT, 0.0f);
		SmartDashboard::PutNumber(SMART_DISC1_GREEN_LIGHT, 0.0f);
		SmartDashboard::PutNumber(SMART_DISC1_RED_LIGHT, 0.0f);
		SmartDashboard::PutNumber(SMART_DISC2_GREEN_LIGHT, 0.0f);
		SmartDashboard::PutNumber(SMART_DISC2_RED_LIGHT, 0.0f);
		SmartDashboard::PutNumber(SMART_DISC3_GREEN_LIGHT, 0.0f);
		SmartDashboard::PutNumber(SMART_DISC3_RED_LIGHT, 0.0f);
		SmartDashboard::PutNumber(SMART_DISC4_GREEN_LIGHT, 0.0f);
		SmartDashboard::PutNumber(SMART_DISC4_RED_LIGHT, 0.0f);
	}

	// set default modes
	_climberMode = false; // shooter mode
	_feedMode = false; // feed button not pressed
	_storageBroken = false; // nothing boroken

	// force lights to be updated by initializing
	// the previous state with every light being on.
	// every light would never be on, so this will
	// force the lights to be updated the first time
	memset(_i2cRegisterValue, 0xFF, sizeof(_i2cRegisterValue));

	ResetLights();
	ResetFlash(); // initialize flash variables
}

WeejusLights::~WeejusLights() {
}

void WeejusLights::forceOn(bool on) {
	_forceAllOn = on;
}

void WeejusLights::forwardToggle(bool climber) {
	_climberMode = climber;

	ResetFlash(); // mode change, so reset flash
}

void WeejusLights::feedToggle(bool feed) {
	_feedMode = feed;
}

void WeejusLights::lightRave(RAVE_PATTERN pattern) {
	if (pattern != _ravePattern) {
		// changing pattern, so reset flash variables
		ResetFlash();

		_ravePattern = pattern; // set after ResetFlash()
	}
}

void WeejusLights::setStorageBroken(bool broken) {
	_storageBroken = broken;
}

void WeejusLights::process() {
	if (_ravePattern == raveDEPLOY_FAIL) {
		processDeployFail();
	} else {
		if (!_climberMode) {
			processShooter();
		} else {
			// lights are in climber mode
			processClimber();
		}
	}

	UpdateLights();
}

// set light variables for when in Shooter mode
void WeejusLights::processShooter() {
	int numberOfDiscs;
	bool isAtAngle;

	// put all lights at a known state
	ResetLights();

	_shooterDirectionLight = true;

	// only the led ring
	_feedGreenLight = (_feedMode && _shooter->getCanFeed());

	// led ring and spotter light
	//_feedGreenLight = (!_feedMode || _shooter->getCanFeed());
	_targetLight = !_feedMode;

	//	printf("canfeed: %i | feedmode: %i\n", _shooter->getCanFeed(), _feedMode);
	_feedRedLight = (_feedMode && !_shooter->getCanFeed());

	numberOfDiscs = _shooter->getNumDiscs();
	isAtAngle = _shooter->getIsAtAngle();
	_disc1GreenLight = (isAtAngle && (numberOfDiscs > 0) && !_storageBroken);
	_disc1RedLight = (!isAtAngle && (numberOfDiscs > 0) && !_storageBroken);
	_disc2GreenLight = (isAtAngle && (numberOfDiscs > 1) && !_storageBroken);
	_disc2RedLight = (!isAtAngle && (numberOfDiscs > 1) && !_storageBroken);
	_disc3GreenLight = (isAtAngle && (numberOfDiscs > 2) && !_storageBroken);
	_disc3RedLight = (!isAtAngle && (numberOfDiscs > 2) && !_storageBroken);
	_disc4GreenLight = (isAtAngle && (numberOfDiscs > 3) && !_storageBroken);
	_disc4RedLight = (!isAtAngle && (numberOfDiscs > 3) && !_storageBroken);
}

// set light variables for when in Climber mode
void WeejusLights::processClimber() {
	double currentTime;

	// put all lights at a known state
	ResetLights();

	_climberDirectionLight = true;

	_disc1GreenLight = _climber->getEngaged();
	_disc1RedLight = !_climber->getEngaged();
	_disc2GreenLight = _climber->getAligned();
	_disc2RedLight = !_climber->getAligned();
	_disc3GreenLight = _climber->getLevelAchieved();
	_disc3RedLight = !_climber->getLevelAchieved();

	// time to switch flash states?
	currentTime = GetTime();
	if ((_lastTime + LEVEL_FLASH) > currentTime) {
		_lastTime = currentTime;
		if (_flashState == 0) {
			_flashState = 1; // lights off
		} else {
			_flashState = 0; // lights on
		}
	}

	// if flash state wants lights off, turn them off
	if (_flashState != 0) {
		if (_climber->getLevelAchieved() > 0) {
			_disc1GreenLight = false;
			_disc1RedLight = false;
		}
		if (_climber->getLevelAchieved() > 1) {
			_disc2GreenLight = false;
			_disc2RedLight = false;
		}
		if (_climber->getLevelAchieved() > 2) {
			_disc3GreenLight = false;
			_disc3RedLight = false;
		}
	}
}

// flashing lights because deploy failed
void WeejusLights::processDeployFail() {
	double currentTime;

	// put all lights at a known state
	ResetLights();

	// turn on lights we want based on the state
	switch (_flashState) {
	case 0:
		_shooterDirectionLight = true;
		_disc1GreenLight = true;
		_disc3RedLight = true;
		break;
	case 1:
		_climberDirectionLight = true;
		_disc2RedLight = true;
		_disc4GreenLight = true;
		break;
	case 2:
		_shooterDirectionLight = true;
		_disc2GreenLight = true;
		_disc3RedLight = true;
		break;
	case 3:
		_climberDirectionLight = true;
		_disc1GreenLight = true;
		_disc4RedLight = true;
		break;
	case 4:
		_shooterDirectionLight = true;
		_disc1RedLight = true;
		_disc3GreenLight = true;
		break;
	case 5:
		_climberDirectionLight = true;
		_disc2GreenLight = true;
		_disc4RedLight = true;
		break;
	case 6:
		_shooterDirectionLight = true;
		_disc1RedLight = true;
		_disc4GreenLight = true;
		break;
	case 7:
		_climberDirectionLight = true;
		_disc2GreenLight = true;
		_disc3RedLight = true;
		break;
	case 8:
		ResetLights(); // all done, go back to normal lights
		ResetFlash();
		return; // don't increment state, get out of here
	}

	// time to switch flash states?
	currentTime = GetTime();
	if ((_lastTime + DEPLOYFAIL_FLASH) > currentTime) {
		_lastTime = currentTime;
		_flashState++;
	}
}

// resets all of the lights to the off position
void WeejusLights::ResetLights() {
	_shooterDirectionLight = false;
	_climberDirectionLight = false;
	_feedGreenLight = false;
	_feedRedLight = false;
	_disc1GreenLight = false;
	_disc1RedLight = false;
	_disc2GreenLight = false;
	_disc2RedLight = false;
	_disc3GreenLight = false;
	_disc3RedLight = false;
	_disc4GreenLight = false;
	_disc4RedLight = false;
	_targetLight = false;
}

// reset flash state variables
void WeejusLights::ResetFlash() {
	_lastTime = GetTime();
	_flashState = 0;
	_ravePattern = raveOFF;
}

// turns lights on/off if the state has changed
void WeejusLights::UpdateLights() {
	UINT8 i2cRegisterValue[I2C_NUMBER_BANKS];
	int i;

	// initialize to lights off
	memset(i2cRegisterValue, 0x00, sizeof(i2cRegisterValue));

	// for each light that is On, set it's bit
	if (_shooterDirectionLight) {
		i2cRegisterValue[I2C_SHOOTERDIRECTION_BANK] |= I2C_SHOOTERDIRECTION_GP;
	}
	if (_climberDirectionLight) {
		i2cRegisterValue[I2C_CLIMBERDIRECTION_BANK] |= I2C_CLIMBERDIRECTION_GP;
	}
	if (_feedGreenLight) {
		i2cRegisterValue[I2C_FEEDGREEN_BANK] |= I2C_FEEDGREEN_GP;
	}
	if (_feedRedLight) {
		i2cRegisterValue[I2C_FEEDRED_BANK] |= I2C_FEEDRED_GP;
	}
	if (_disc1GreenLight) {
		i2cRegisterValue[I2C_DISC1GREEN_BANK] |= I2C_DISC1GREEN_GP;
	}
	if (_disc1RedLight) {
		i2cRegisterValue[I2C_DISC1RED_BANK] |= I2C_DISC1RED_GP;
	}
	if (_disc2GreenLight) {
		i2cRegisterValue[I2C_DISC2GREEN_BANK] |= I2C_DISC2GREEN_GP;
	}
	if (_disc2RedLight) {
		i2cRegisterValue[I2C_DISC2RED_BANK] |= I2C_DISC2RED_GP;
	}
	if (_disc3GreenLight) {
		i2cRegisterValue[I2C_DISC3GREEN_BANK] |= I2C_DISC3GREEN_GP;
	}
	if (_disc3RedLight) {
		i2cRegisterValue[I2C_DISC3RED_BANK] |= I2C_DISC3RED_GP;
	}
	if (_disc4GreenLight) {
		i2cRegisterValue[I2C_DISC4GREEN_BANK] |= I2C_DISC4GREEN_GP;
	}
	if (_disc4RedLight) {
		i2cRegisterValue[I2C_DISC4RED_BANK] |= I2C_DISC4RED_GP;
	}
	if (_targetLight) {
		i2cRegisterValue[I2C_SHOOTERLIGHT_BANK] |= I2C_SHOOTERLIGHT_GP;
	}

	if (_forceAllOn) {
		memset(i2cRegisterValue, 0xFF, sizeof(i2cRegisterValue));
	}

	// for each register that is different from the last
	// time we set the lights, we need to send to the new
	// GP IO bits to it. we do this by writing to the
	// GPIO_REGISTER (i.e. 0x09) portion of that bank

	for (i = 0; i < I2C_NUMBER_BANKS; i++) {
		// see if GP IO bits are different
		if (_i2cRegisterValue[i] != i2cRegisterValue[i]) {
			_i2cRegisterValue[i] = i2cRegisterValue[i];
			if ((_i2c[i] != NULL) && _i2c[i]->Write((UINT8) 0x09,
					_i2cRegisterValue[i])) {
				//printf("ERROR writing to I2C %i\n", (int)_i2cBankAddress[i]);
			}
		}
	}

	// make sure the Ring light and SmartDashboard are updated
	UpdateLightsRing();
	UpdateLightsSmartDashboard();
}

// turns Camera Ring light on/off if the state has changed
void WeejusLights::UpdateLightsRing() {
	// if not using ring light for feeding, then nothing to do
	if (!_useRingForFeedLight) {
		return;
	}

	//	printf("update light ring RFL: %i | FGL: %i\n", _ringFeedLight, _feedGreenLight);
	if (_ringFeedLight != _feedGreenLight) {
		//printf("")
		_ringFeedLight = _feedGreenLight;
		WeejusTarget::getInstance()->setLight(_ringFeedLight);
	}
}

// turns SmartDashboard lights on/off if the state has changed
// NOTE: if you added anything here, make sure you update the WeejusLights constructor
void WeejusLights::UpdateLightsSmartDashboard() {
	// if not using SmartDashboard, then nothing to do
	if (!_useSmartDashboard) {
		return;
	}

	// only update SmartDashboard if something changed
	if (_shooterDirectionLightSmartDashboard != _shooterDirectionLight) {
		_shooterDirectionLightSmartDashboard = _shooterDirectionLight;
		SmartDashboard::PutNumber(SMART_SHOOTER_DIR_LIGHT,
				(double) _shooterDirectionLightSmartDashboard);
	}
	if (_climberDirectionLightSmartDashboard != _climberDirectionLight) {
		_climberDirectionLightSmartDashboard = _climberDirectionLight;
		SmartDashboard::PutNumber(SMART_CLIMBER_DIR_LIGHT,
				(double) _climberDirectionLightSmartDashboard);
	}
	if (_feedGreenLightSmartDashboard != _feedGreenLight) {
		_feedGreenLightSmartDashboard = _feedGreenLight;
		SmartDashboard::PutNumber(SMART_FEED_GREEN_LIGHT,
				(double) _feedGreenLightSmartDashboard);
	}
	if (_feedRedLightSmartDashboard != _feedRedLight) {
		_feedRedLightSmartDashboard = _feedRedLight;
		SmartDashboard::PutNumber(SMART_FEED_RED_LIGHT,
				(double) _feedRedLightSmartDashboard);
	}
	if (_disc1GreenLightSmartDashboard != _disc1GreenLight) {
		_disc1GreenLightSmartDashboard = _disc1GreenLight;
		SmartDashboard::PutNumber(SMART_DISC1_GREEN_LIGHT,
				(double) _disc1GreenLightSmartDashboard);
	}
	if (_disc1RedLightSmartDashboard != _disc1RedLight) {
		_disc1RedLightSmartDashboard = _disc1RedLight;
		SmartDashboard::PutNumber(SMART_DISC1_RED_LIGHT,
				(double) _disc1RedLightSmartDashboard);
	}
	if (_disc2GreenLightSmartDashboard != _disc2GreenLight) {
		_disc2GreenLightSmartDashboard = _disc2GreenLight;
		SmartDashboard::PutNumber(SMART_DISC2_GREEN_LIGHT,
				(double) _disc2GreenLightSmartDashboard);
	}
	if (_disc2RedLightSmartDashboard != _disc2RedLight) {
		_disc2RedLightSmartDashboard = _disc2RedLight;
		SmartDashboard::PutNumber(SMART_DISC2_RED_LIGHT,
				(double) _disc2RedLightSmartDashboard);
	}
	if (_disc3GreenLightSmartDashboard != _disc3GreenLight) {
		_disc3GreenLightSmartDashboard = _disc3GreenLight;
		SmartDashboard::PutNumber(SMART_DISC3_GREEN_LIGHT,
				(double) _disc3GreenLightSmartDashboard);
	}
	if (_disc3RedLightSmartDashboard != _disc3RedLight) {
		_disc3RedLightSmartDashboard = _disc3RedLight;
		SmartDashboard::PutNumber(SMART_DISC3_RED_LIGHT,
				(double) _disc3RedLightSmartDashboard);
	}
	if (_disc4GreenLightSmartDashboard != _disc4GreenLight) {
		_disc4GreenLightSmartDashboard = _disc4GreenLight;
		SmartDashboard::PutNumber(SMART_DISC4_GREEN_LIGHT,
				(double) _disc4GreenLightSmartDashboard);
	}
	if (_disc4RedLightSmartDashboard != _disc4RedLight) {
		_disc4RedLightSmartDashboard = _disc4RedLight;
		SmartDashboard::PutNumber(SMART_DISC4_RED_LIGHT,
				(double) _disc4RedLightSmartDashboard);
	}

	// NOTE: if you added anything here, make sure you update the WeejusLights constructor
}
