#ifndef _LIGHTS_H_
#define _LIGHTS_H_
#include "WPILib.h"
#include "WeejusShooter.h"
#include "WeejusClimber.h"
#include "WeejusTarget.h"
#include "DSMap.h"

// if ok to use I2C (i.e. light hardward in place), set to 1
#define USE_I2C			1

// duration of the flash for level achieved
#define LEVEL_FLASH		0.25f

// duration for each state of the rave.
// we have 8 steps, so .25 sec gives a total rave of 2 seconds
#define DEPLOYFAIL_FLASH	0.25f

class WeejusLights {
public:
	/*
	 * Rave you want to experience
	 * DEPLOY_FAIL: when unable to deploy
	 */
	typedef enum RAVE_PATTERN {
		raveOFF, raveDEPLOY_FAIL, raveFUN1, raveFUN2
	};

	//DO NOT CALL THIS, other then in Controls.cpp
	WeejusLights(bool useSmartDashboard, bool useRingForFeedLight);
	~WeejusLights();

	/*
	 * Method: forwardToggle();
	 * 
	 * Inputs: bool climber
	 * 			true		climber
	 * 			false		shooter
	 * Outputs: Null
	 * 
	 * Remarks: 
	 * Indicated which side of the robot is the front
	 * Called by WeejusControls when 'reverse' button changes
	 */
	void forwardToggle(bool climber);

	/** Force all lights to on if true otherwise
	 *  control lights as per other functions herein
	 */
	void forceOn(bool on);

	/*
	 * Method: feedToggle();
	 * 
	 * Inputs: bool feed
	 * 			true  if in Feed mode (button is pressed)
	 * 			false if not in Feed mode (button not pressed)
	 * Outputs: Null
	 * 
	 * Remarks: 
	 * Turns Feed light on/off
	 * Called by WeejusControls when Feed mode button changes
	 */
	void feedToggle(bool feed);

	/*
	 * Method: lightRave();
	 * 
	 * Inputs: RAVE_PATTERN to execute
	 * Outputs: Null
	 * 
	 * Remarks: 
	 * Turns lights on/off based on the given pattern
	 * Called by WeejusControls
	 */
	void lightRave(RAVE_PATTERN pattern);

	/*
	 * Method: setStorageBroken(bool);
	 * 
	 * Inputs:
	 * 			bool broken: bool representing whether any storage sensors are broken
	 * Outputs: Null
	 * 
	 * Remarks:
	 * called by WeejusControls when a broken switch is flipped
	 */
	void setStorageBroken(bool broken);

	/*
	 * Method: process();
	 * 
	 * Inputs: Null
	 * Outputs: Null
	 * 
	 * Remarks:
	 * Updates all lights and internal status variables
	 * MUST be called at the END of each loop to continue operation of the climber
	 */
	void process();

private:
	void processShooter();
	void processClimber();
	void processDeployFail();
	void ResetLights();
	void ResetFlash();
	void UpdateLights();
	void UpdateLightsRing();
	void UpdateLightsSmartDashboard();

	WeejusShooter *_shooter;
	WeejusClimber *_climber;
	bool _useSmartDashboard;
	bool _useRingForFeedLight;

	// modes and broken switches
	bool _climberMode;
	bool _feedMode;
	bool _storageBroken;

	// I2C object and last set of values that was sent
	DigitalModule *_digitalModule;
	I2C *_i2c[I2C_NUMBER_BANKS];
	UINT8 _i2cRegisterValue[I2C_NUMBER_BANKS];

	bool _forceAllOn;

	// state for each light
	bool _shooterDirectionLight;
	bool _climberDirectionLight;
	bool _feedGreenLight;
	bool _feedRedLight;
	bool _disc1GreenLight;
	bool _disc1RedLight;
	bool _disc2GreenLight;
	bool _disc2RedLight;
	bool _disc3GreenLight;
	bool _disc3RedLight;
	bool _disc4GreenLight;
	bool _disc4RedLight;

	// state for camera ring light, when being used for Feed light
	bool _ringFeedLight;
	bool _targetLight;

	// state for each light on SmartDashboard
	bool _shooterDirectionLightSmartDashboard;
	bool _climberDirectionLightSmartDashboard;
	bool _feedGreenLightSmartDashboard;
	bool _feedRedLightSmartDashboard;
	bool _disc1GreenLightSmartDashboard;
	bool _disc1RedLightSmartDashboard;
	bool _disc2GreenLightSmartDashboard;
	bool _disc2RedLightSmartDashboard;
	bool _disc3GreenLightSmartDashboard;
	bool _disc3RedLightSmartDashboard;
	bool _disc4GreenLightSmartDashboard;
	bool _disc4RedLightSmartDashboard;

	// flash state variables
	double _lastTime;
	int _flashState;
	RAVE_PATTERN _ravePattern;
};
#endif
