#include "WeejusShooter.h"
#include <stdio.h>
#include <math.h>

// use the Byers (c) speed controller
#define USE_SEMIBANGBANG_SPEED_CONTROL 1
#define PID_CONTROL 0

//initialization, sets by functions called by other classes, etc
#define SHOOTER_DEBUG_GENERIC 0
//autonomous "shoot all" state machine
#define SHOOTER_DEBUG_AUTO 0
//teleoperated automatic shooting and indexing state machine
//TODO turn off when deploying final code
#define SHOOTER_DEBUG_TELEOP 1

// will backup the indexer right after waiting for it to drop
// and beofre moving to feed sensor
#define SHOOTER_INDEX_BACKUP 1

// force the indexer to see the Feed/2nd sensor before
// continuing to index to Load/1st sensor
#define SHOOTER_FORCE_AT_FEED	1

WeejusShooter* WeejusShooter::_theShooter = NULL;

WeejusShooter* WeejusShooter::getInstance() {
	if (_theShooter == NULL) {
		_theShooter = new WeejusShooter();
	}
	return _theShooter;
}

WeejusShooter::WeejusShooter() {
	//generic member vars
	sprintf(_dbgBuf, "weejus");
	_wntAngle = 0.0f;
	_numDiscs = 0;
	_discErrorCount = 0;
	_waitTimeStart = 0.0f;
	_waitShootStart = 0.0f;
	_speedWaitTimeOvr = 0.0f;
	_shootSpeed = 0.0f;

	_statShooting = false;
	_statShootDone = true;
	_shootState = sStart;
	_reInitializeShootState = false;
	_discInShooter = false;

	_waitIndexStart = 0.0f;
	_reInitializeIndexState = false;
	_discInFeeder = false;

	_statShootOverride = false;
	_statAutoOverride = false;
	_statFeedMode = false;
	_statCanFeed = false;
	_statMoveIndexer = false;

	_sensorDebug = false;

	//broken switches
	_brkAngle = false;
#if FORCE_ENCODER_BROKEN
	_brkEncoder = true; // can only work in broken state (see setEncoderBroken() if this changes)
#else
	_brkEncoder = false;
#endif
	_brkStorage = false;
	_brkDead = false;

	_log = new UDPLogger();
	// 1140 is legal outgoing UDP port during competition; 1130 is legal ingoing
	// 1511 is the graphlog software
	_log->init(1140);

	//initialize all needed sensors
	initSensors();

	//initialize all needed motors/outputs
	initOutputs();

	//turn on logger
	_log->setEnabled(SHOOTER_LOG);
}

WeejusShooter::~WeejusShooter() {
	delete _sensAngle;
	delete _sensSpeed;
	delete _sensLoaded;
	delete _sensAlignLoad;
	delete _sensAlignFeed;
	delete _sensFed;

	delete _mtrAngle;
	delete _mtrShoot;
	delete _mtrIndexer;
	delete _srvGuard;
	delete _srvShoot;
}

void WeejusShooter::moveToAngle(float angle) {
	//set the wanted angle(or direction if broken)
	if (angle < -1.0f) {
		angle = -1.0f;
	} else if (angle > 1.0f) {
		angle = 1.0f;
	}
	_wntAngle = angle;
#if SHOOTER_DEBUG_GENERIC
	printf("[GEN] Wanted Angle: %f\n", _wntAngle);
#endif
}

bool WeejusShooter::getIsAtAngle() {
	return checkShooterAngle(
			linearRangeScale(_wntAngle, -1.0f, 1.0f, ANGLEPOT_MIN, ANGLEPOT_MAX));
}

float WeejusShooter::calcShooterAngle() {

	if (_brkAngle)
		return -1;

	//convert pot value to raw slider value
	return linearRangeScale(_sensAngle->GetAverageVoltage(), ANGLEPOT_MIN,
			ANGLEPOT_MAX, -1.0f, 1.0f);
}

bool WeejusShooter::getClimbReady() {
	if (_brkAngle || _brkDead)
		return 1;
	else
		return (calcShooterAngle() >= ANGLE_CLIMB);
}

void WeejusShooter::setNumDiscs(int discs) {
	_numDiscs = discs;
	if (_numDiscs < 0)
		_numDiscs = 0;
	if (_numDiscs > 4)
		_numDiscs = 4;

#if SHOOTER_DEBUG_GENERIC
	printf("[GEN] Set # Discs To: %i\n", _numDiscs);
#endif
}

void WeejusShooter::initialize(bool autonoumous) {
	_discErrorCount = 0;
	_speedWaitTimeOvr = 0.0f; // restart override start time
	_timeOfLastShot = 0.0f;

	_pidIntegral = 0.0;
	_pidLastError = 0.0;

	_statShooting = false;
	_reInitializeShootState = true;
	shootOverride(_statShootOverride); // make sure the time is reset

	_statMoveIndexer = false;

	// indicate the state machine needs to exit any wait loops.
	// be we want them to finish any indexing so if we stop in
	// the middle of that and then have a mess on our hands
	_reInitializeIndexState = true;

	// reset out flags based on the current state of the sensors
	_discInShooter = (_sensLoaded->Get() == 0);
	_discInFeeder = (_sensFed->Get() == 0);

	_mtrAngle->Set(0.0f);
}

void WeejusShooter::process() {
	//process any manual stuff
	if (_brkDead) {
		//printf("dead\n");
		//make sure all motors are stopped
		_mtrAngle->Set(0.0f);
		_mtrIndexer->Set(0.0f);

		stopShooterWheel();
		return;
	}

	processAngleController();
	processIndexLoop();
	processShootLoop();

	// since we are ignoring the disc count, we will
	// allow feeding if Waiting For a Disc
	//printf("at angle: %i\n", checkShooterAngle(ANGLE_FEED));
	float angScl = linearRangeScale(ANGLE_FEED, -1.0f, 1.0f, ANGLEPOT_MIN,
			ANGLEPOT_MAX);
	_statCanFeed = checkShooterAngle(angScl);

	//#if SHOOTER_LOG
	//	double v = _sensSpeed->PIDGet();
	//	_log->log("%f,%f,%f,%f\n", (double)_shootSpeed-v, (double)_shootSpeed, (double)v, (double)_shootSpeed);
	//#endif
}

/************/
/*AUTONOMOUS*/
/************/

void WeejusShooter::shootAll() {
	_statShooting = true;
	_statShootDone = false;
#if SHOOTER_DEBUG_GENERIC
	printf("[GEN] Start Shoot All\n");
#endif
}

bool WeejusShooter::shootDone() {
	return _statShootDone;
}

/**************/
/*TELEOPERATED*/
/**************/

int WeejusShooter::getNumDiscs() {
	return _numDiscs;
}

bool WeejusShooter::getCanFeed() {
	return _statCanFeed;
}

void WeejusShooter::shootOverride(bool override) {
	_statShootOverride = override;
}

void WeejusShooter::autoWheelOn(bool on) {
	_statAutoOverride = on;
}

void WeejusShooter::startShoot() {
	_statShooting = true;
	_statShootDone = false;
#if SHOOTER_DEBUG_GENERIC
	printf("[GEN] Start Shooting\n");
#endif
}

void WeejusShooter::stopShoot() {
	_statAutoOverride = false;

	_statShooting = false;
	_reInitializeShootState = true;
	_reInitializeIndexState = true; // do not re-start if in the middle of indexing
#if SHOOTER_DEBUG_GENERIC
	printf("[GEN] Stop Shooting\n");
#endif
}

void WeejusShooter::feedMode(bool on) {
	if (_statFeedMode != on) {
#if SHOOTER_DEBUG_TELEOP
		printf("[TELEOP] Feed mode changed: %d\n", on);
#endif	
		_discErrorCount = 0; // force 'error' loop
		_speedWaitTimeOvr = 0.0f;// restart override start time
	}
	//set feed mode
	_statFeedMode = on;
}

/****************/
/*MANUEL METHODS*/
/****************/

void WeejusShooter::setAngleBroken(bool broken) {
	_brkAngle = broken;
}

void WeejusShooter::setEncoderBroken(bool broken) {
	if (broken != _brkEncoder) {
		printf("Shooter encoder broken state chnage: %d\n", broken);
		stopShooterWheel();
	}
	_brkEncoder = broken;

#if FORCE_ENCODER_BROKEN
	// right now we can ONLY work in broken mode, so leave it set to TRUE
	// NOTE: if this gets removed, then change the constructor so it
	//       will default _brkEncoder to false
	_brkEncoder = true;
#endif
}

void WeejusShooter::setStorageBroken(bool broken) {
	_brkStorage = broken;
}

void WeejusShooter::setDead(bool dead) {
	_brkDead = dead;
}

void WeejusShooter::setSensorDebug(bool debug) {
	_sensorDebug = debug;
}

void WeejusShooter::setSpeed(float power) {
	//make sure speed is within the bounds
	if (power > 1.0f)
		_shootSpeed = 1.0f;
	else if (power < 0.0f)
		_shootSpeed = 0.0f;
	else
		_shootSpeed = power;
#if SHOOTER_DEBUG_GENERIC
	printf("[GEN] Set Speed to: %f\n", _shootSpeed);
#endif
}

void WeejusShooter::turnIndexer() {
	_statMoveIndexer = true;
}

char* WeejusShooter::getDebugLine(int line) {
	//return debug line based on wanted line
	switch (line) {
	default:
		snprintf(_dbgBuf, sizeof(_dbgBuf), "invalid weejus");
		break;
	case 1:
		snprintf(_dbgBuf, sizeof(_dbgBuf), "Load: %i", _sensLoaded->Get());
		break;
	case 2:
		snprintf(_dbgBuf, sizeof(_dbgBuf), "SG: %1.3f|SF: %1.3f",
				(double) _srvGuard->Get(), (double) _srvShoot->Get());
		break;
	case 3:
		snprintf(_dbgBuf, sizeof(_dbgBuf), "CnFeed: %i|IsShtg: %i",
				_statCanFeed, _statShooting);
		break;
	case 4:
		_sensSpeed->PIDGet();
		snprintf(_dbgBuf, sizeof(_dbgBuf), "RPM:%04i|Nrm:%f",
				(int) (_sensSpeed->getRevPerSec() * 60.0f),
				(double) _sensSpeed->getLastSpeed());
		break;
	case 5:
		snprintf(
				_dbgBuf,
				sizeof(_dbgBuf),
				"WA: %f|SS:%f",
				(double) linearRangeScale(_wntAngle, -1.0f, 1.0f, ANGLEPOT_MIN,
						ANGLEPOT_MAX), (double) _shootSpeed);
		break;
	case 6:
		snprintf(_dbgBuf, sizeof(_dbgBuf), "ND: %d|DE: %d|ICur %1.2f",
				_numDiscs, _discErrorCount,
				(double) _mtrIndexer->GetOutputCurrent());
		break;
	}
	return _dbgBuf;
}

void WeejusShooter::initSensors() {
#if SHOOTER_DEBUG_GENERIC
	printf("[GEN] Start Init Sensors...\n");
#endif
	_sensAngle = new AnalogChannel(ANA_LEFT_SHOOTER_TILT_SENSOR);
	// 2^9 samples; this needs 512+ samples/5 msec, close to
	// analog module sampling max for 3 analog channels
	//_sensAngle->SetAverageBits(6);
	//_sensAngle->SetOversampleBits(2);
	_sensSpeed = new MonoEncoder(DIG_SHOOTER_WHEEL_ENCODER_A, SHOOTER_MAX_RPS);
	_sensLoaded = new DigitalInput(DIG_SHOOTER_READY_TO_SHOOT);
	_sensAlignLoad = new DigitalInput(DIG_SHOOTER_LOADER_ALIGN);
	_sensAlignFeed = new DigitalInput(DIG_SHOOTER_FEED_ALIGN);
	_sensFed = new DigitalInput(DIG_SHOOTER_DISC_1_PRESENT);
#if SHOOTER_DEBUG_GENERIC
	printf("[GEN] Done\n");
#endif
}
void WeejusShooter::initOutputs() {
#if SHOOTER_DEBUG_GENERIC
	printf("[GEN] Start Init Outputs...\n");
#endif
	_mtrAngle = new WeejusCANJaguar(CANID_SHOOTER_TILT_MOTOR);
	_mtrShoot = new WeejusCANJaguar(CANID_SHOOTER_WHEEL_MOTOR);
	_mtrShoot->ConfigNeutralMode(WeejusCANJaguar::kNeutralMode_Coast);
	_mtrShoot->SetSafetyEnabled(false);

	_mtrIndexer = new WeejusCANJaguar(CANID_SHOOTER_CAROUSEL_MOTOR);
	_mtrIndexer->ConfigNeutralMode(WeejusCANJaguar::kNeutralMode_Brake);
	_mtrIndexer->SetSafetyEnabled(false);

	_srvGuard = new Servo(PWM_SHOOTER_LOAD_BLOCKER_SEVO);
	_srvShoot = new Servo(PWM_SHOOTER_LOAD_PUSHER_SERVO);
#if SHOOTER_DEBUG_GENERIC
	printf("[GEN] Done\n");
#endif
}

bool WeejusShooter::checkShooterAngle(float angle) {
	float calcAngle = _sensAngle->GetAverageVoltage();//calcShooterAngle();
	if (calcAngle >= angle - ANGLE_TOLERANCE && calcAngle <= angle
			+ ANGLE_TOLERANCE)
		return true;

	return false;
}

float WeejusShooter::linearRangeScale(float v, float min, float max,
		float newMin, float newMax) {
	float oldr = (max - min);
	float newr = (newMax - newMin);
	return (((v - min) * newr) / oldr) + newMin;
}

void WeejusShooter::processAngleController() {
	float potV = _sensAngle->GetAverageVoltage();
	//printf("process angle\n");
	if (!_brkAngle) {
		//calculate wanted pot value by scaling the slider range to the pot range between the limits
		float angScl = linearRangeScale(_wntAngle, -1.0f, 1.0f, ANGLEPOT_MIN,
				ANGLEPOT_MAX);
		//check limits before anything else
		if ((potV > angScl) && potV <= ANGLEPOT_MIN) {
			//printf("nope, at min limit\n");
			_mtrAngle->Set(0.0f);
			return;
		} else if ((potV < angScl) && potV >= ANGLEPOT_MAX) {
			//printf("nope, at max limit wnt: %f | pot: %f\n", (double)_wntAngle, (double)potV);
			_mtrAngle->Set(0.0f);
			return;
		}

		if (checkShooterAngle(angScl)) {
			//do nothing if we are at the desired angle
			_mtrAngle->Set(0.0f);
			//printf("at angle\n");
		} else if (potV < angScl) {
			//move up if we are below the desired angle
			//motor is reversed
			if ((angScl - potV) < SLOW_ANGLE_TOLERANCE) {
				_mtrAngle->Set(ANGLE_SPEED_MED);
			} else {
				_mtrAngle->Set(ANGLE_SPEED_FAST);
			}
			//printf("up\n");
		} else if (potV > angScl) {
			//move down if we are above the desired angle
			//motor is reversed
			if ((potV - angScl) < SLOW_ANGLE_TOLERANCE) {
				_mtrAngle->Set(-ANGLE_SPEED_MED);
			} else {
				_mtrAngle->Set(-ANGLE_SPEED_FAST);
			}
			//printf("down\n");
		}
	} else {
		//check dead zone
		//printf("angle broken\n");
		if (_wntAngle >= -ANGLE_DEAD_ZONE && _wntAngle <= ANGLE_DEAD_ZONE) {
			_mtrAngle->Set(0.0f);
			return;
		}

		//calculate direction and directly set the motors b/c pot is broken
		//will move slower b/c we can't use limits since the pot is broken
		if (_wntAngle < 0.0f) {
			//move down at predefined speed if wanted angle is negative
			//we want to move the motor slower
			_mtrAngle->Set(-ANGLE_SPEED_BROKEN);
		} else if (_wntAngle > 0.0f) {
			//move up at predefined speed if wanted angle is positive
			//we want to move the motor slower
			_mtrAngle->Set(ANGLE_SPEED_BROKEN);
		}
	}
}
void WeejusShooter::processIndexLoop() {
	double spd;
	if (_brkStorage) {
		//if the storage is broken, we want to just move the indexer at the predefined speed
		if (_statMoveIndexer) {
			spd = DEF_INDEXER_SPEED_MAN;
			//set _statMoveIndexer to false so movement will stop until
			//moveIndexer is called again
			_statMoveIndexer = false;
		} else {
			spd = 0.0f;
		}
	} else {
		// sensor low ->  blocked, sensor high -> clear
		_discInShooter = (_sensLoaded->Get() == 0);

		if (_discInShooter) {
			spd = DEF_INDEXER_SPEED_FULL;
		} else {
			spd = DEF_INDEXER_SPEED_EMPTY;
		}
	}
	_mtrIndexer->Set(spd);
}

// TODO finish PID loop
void WeejusShooter::runShooterWheel() {
	double output;
	//if we are continuously shooting
	if (_brkEncoder) {
		//if broken, set directly
		output = _shootSpeed;
	} else {
#if USE_SEMIBANGBANG_SPEED_CONTROL
		//Byers version
		double spd = _sensSpeed->PIDGet() + VALUE_TO_ENCODER_OFFSET;
		if (spd > (_shootSpeed * 0.9)) {
			output = _shootSpeed;
		} else {
			output = 1.0f;
		}
#elif PID_CONTROL		
		double spd = _sensSpeed->PIDGet();
		double currError = _shootSpeed - spd;// positive on undershoot

		_pidIntegral += currError;
		double diffError = currError - _pidLastError;

		output = (PID_F*_shootSpeed) + (PID_P * currError) + (PID_D * diffError) + (PID_I * _pidIntegral);

		_pidLastError = currError;
#else		
		if (_sensSpeed->PIDGet() > _shootSpeed) {
			output = 0.0f;
		} else {
			output = 1.0f;
		}
#endif

	}

	_mtrShoot->Set(-output);
#if SHOOTER_LOG
	_log->log("S:%f,%f,%f,%f\n", _shootSpeed, output, _sensSpeed->PIDGet(),
			_sensSpeed->PIDGet() - output);
#endif

	// set the override time if not already set
	if (_speedWaitTimeOvr == 0.0f) {
		_speedWaitTimeOvr = GetTime();
	}
	if (_timeOfLastShot == 0.0f) {
		_timeOfLastShot = GetTime();
	}

}

void WeejusShooter::stopShooterWheel() {
	_waitShootStart = 0.0; // need to wait again when we re-start the wheel
	_speedWaitTimeOvr = 0.0;// need to wait again when we re-start the wheel
	_timeOfLastShot = 0.0;
	_mtrShoot->Set(0.0f);
	_pidIntegral = 0.0;
	_pidLastError = 0.0;
}

void WeejusShooter::processShootLoop() {
	// if in shooting mode AND wheel override is ON,
	// then make sure wheel continues to spin
	if (!_statFeedMode && (_statShootOverride || _statAutoOverride)) {
		runShooterWheel();
	}

	switch (_shootState) {
	case sStart:
		_reInitializeShootState = false;

		// if not being asked to shoot, then nothing to do
		if (!_statShooting) {
			//make sure the guard and flapper in the correct position
			_srvGuard->Set(GUARD_UP_POS);
			_srvShoot->Set(SERVO_LOAD_POS);
			_statShootDone = true;

			// if in feed mode OR wheel override is OFF, then
			// we need to turn off the wheel
			if (_statFeedMode || (!_statShootOverride && !_statAutoOverride)) {
				stopShooterWheel();
			}
			break;
		}

		// if storage is broken, then we will shoot whenever asked.
		// so if storage is not broken, then only shoot when indexer is ready
		if (!_brkStorage) {
			// if indexer is not ready, then wait for next loop
			if (!_discInShooter)
				break;

			// indexer is ready, so tell it we are now shooting
		}

		if (SHOOTER_DEBUG_TELEOP)
			printf("[TELEOP] Speed up wheel... %.4f\n", (double) GetTime());

		runShooterWheel(); // get wheel spinning

		//if BROKEN we will wait a certian amount of time before assuming we are at speed
		//if NOT BROKEN this will be used for the PID timeout override
		// NOTE: only reset if wheel not already running
		if (_waitShootStart == 0.0f) {
			if (SHOOTER_DEBUG_TELEOP)
				printf(
						"[TELEOP] Wheel was not already running (unless override: %d)...\n",
						_statShootOverride);

			_waitShootStart = GetTime();
		}

		_shootState = sSpeedWait;
		break;

	case sSpeedWait:
		// if need to reset state machince, it is ok to do it here
		// because we have not moved any guards yet
		if (_reInitializeShootState) {
			if (SHOOTER_DEBUG_TELEOP)
				printf("[TELEOP] Reset shooter state machine... %.4f\n",
						(double) GetTime());

			_shootState = sStart;
			stopShooterWheel(); // easiest way to reset stuff
			break;
		}

		//wait for the wheel to be up to speed

		if (_brkEncoder) {
			//we will continue shooting if
			//wheel speed up time has passed (normal or override)
			//get actual speed of the wheel
			//measured speed was calculated to be consistently off with the sent value by 0.05 

			// simple time based
			if ((GetTime() - _waitShootStart >= SPEED_WAIT_TIME)
					|| ((_statShootOverride || _statAutoOverride)
							&& ((GetTime() - _speedWaitTimeOvr)
									>= SPEEDPID_WAIT_OVERRIDE))) {
				if (SHOOTER_DEBUG_TELEOP) {
					printf("[TELEOP] Done waiting for time %.4f\n",
							(double) GetTime());
					printf("rpm: %i|nrm: %f\n",
							(int) _sensSpeed->getRevPerSec() * 60,
							_sensSpeed->getLastSpeed());
				}
				_shootState = sDropGuard;
			}
		} else {
			// encoder works

			bool timedOutP = (GetTime() - _timeOfLastShot >= 5.0);
#if USE_SEMIBANGBANG_SPEED_CONTROL
			if (((_sensSpeed->PIDGet() + VALUE_TO_ENCODER_OFFSET)
					>= _shootSpeed) || timedOutP)
#else				
			// simple
			if ( (_sensSpeed->PIDGet() >= _shootSpeed) ||
					timedOutP )
#endif

			{
				if (SHOOTER_DEBUG_TELEOP) {
					printf("[TELEOP] Speed on target %.4f\n",
							(double) GetTime());
					printf("rpm: %i|nrm: %f\n",
							(int) _sensSpeed->getRevPerSec() * 60,
							_sensSpeed->getLastSpeed());
					if (timedOutP) {
						printf("Timeout: shooting anyway %.4f\n",
								GetTime() - _timeOfLastShot);
					}
				}

				_shootState = sDropGuard;
			}
		}

		runShooterWheel(); // keep wheel spinning
		break;

	case sDropGuard:
		// no need to look at the reset state machine flag because
		// this state happens just once, let next state deal with it

		if (SHOOTER_DEBUG_TELEOP)
			printf("[TELEOP] Dropping Guard... %.4f\n", (double) GetTime());

		//drop the guard
		_srvGuard->Set(GUARD_DOWN_POS);
		//start timer to wait for guard to be completely down
		_waitTimeStart = GetTime();
		_shootState = sShoot;
		runShooterWheel(); // keep wheel spinning
		break;

	case sShoot:
		// if we need to reset state machince, we have to take care.
		// at this point we are in the middle of shooting, so we 
		// want to make sure that goes through that entire cycle.
		// if we don't do that, we could end up with a disc hung up
		// on one of the guards as it might not have fully shot.
		// so we will re-run the shot sequence to be sure the wheel
		// is up to speed
		if (_reInitializeShootState) {
			if (SHOOTER_DEBUG_TELEOP)
				printf("[TELEOP] Retry shooting (sShoot)... %.4f\n",
						(double) GetTime());

			_reInitializeShootState = false; // need to force re-shoot
			_shootState = sSpeedWait;
			stopShooterWheel(); // easiest way to reset stuff
			break;
		}

		// if enough time has not past, then we will do nothing but keep the motor powered
		if ((GetTime() - _waitTimeStart) < GUARD_DOWN_WAIT_TIME) {
			runShooterWheel();
			break;
		}

		if (SHOOTER_DEBUG_TELEOP)
			printf("[TELEOP] Shooting... %.4f\n", (double) GetTime());

		_srvShoot->Set(SERVO_SHOOT_POS);

		//start shot wait timer
		_waitTimeStart = GetTime();
		_timeOfLastShot = GetTime();
		_shootState = sWaitShot;
		runShooterWheel(); // keep wheel spinning
		break;

	case sWaitShot:
		// if we need to reset state machince, we have to take care.
		// at this point we are in the middle of shooting, so we 
		// want to make sure that goes through that entire cycle.
		// if we don't do that, we could end up with a disc hung up
		// on one of the guards as it might not have fully shot.
		// so we will re-run the shot sequence to be sure the wheel
		// is up to speed
		if (_reInitializeShootState) {
			if (SHOOTER_DEBUG_TELEOP)
				printf("[TELEOP] Retry shooting (sWaitShoot)... %.4f\n",
						(double) GetTime());

			_reInitializeShootState = false; // need to force re-shoot
			_shootState = sSpeedWait;
			stopShooterWheel(); // easiest way to reset stuff
			break;
		}

		// if enough time has not past, then we will do nothing but keep the motor powered
		if ((GetTime() - _waitTimeStart) < PUSHER_WAIT_TIME) {
			runShooterWheel();
			break;
		}

		if (SHOOTER_DEBUG_TELEOP)
			printf(
					"[TELEOP] Disc shot, wait for pusher to return & disc gone %.4f\n",
					(double) GetTime());

		//reset the guard and flapper
		_srvGuard->Set(GUARD_UP_POS);
		_srvShoot->Set(SERVO_LOAD_POS);

		// no need to wait for wheel because it will
		// take longer to index than it will to re-start
		//_waitShootStart = 0.0f;	// need to wait again because wheel got slowed down		

		// wait for the pusher to clear the space so we can make sure the
		// disc left the shooter
		_shootState = sWaitClear;
		_waitTimeStart = GetTime();
		runShooterWheel(); // keep wheel spinning
		break;

	case sWaitClear:
		// if enough time has not past, then we will do nothing but keep the motor powered
		if ((GetTime() - _waitTimeStart) < CLEAR_WAIT_TIME) {
			runShooterWheel();
			break;
		}

		// pusher should be back and disc cleared. if not, then
		// we need to shoot again to try and clear the shooter
		// before we try to put another disc in there.
		// NOTE: but only if storage is not broken
		if (!_brkStorage && !_sensLoaded->Get()) {
			if (SHOOTER_DEBUG_TELEOP)
				printf(
						"[TELEOP] Shooter is still blocked, so shoot again %.4f\n",
						(double) GetTime());

			_shootState = sStart;
			runShooterWheel(); // keep wheel running
			break;
		}

		// everything is clear, so move on to next disc
		if (SHOOTER_DEBUG_TELEOP)
			printf("[TELEOP] Shooter is clear %.4f\n", (double) GetTime());

		// updating _numDiscs is handled in indexer state machine

		//reset shoot state machine: will shot again if there is a disc
		_shootState = sStart;

		break;
	}
}
