#ifndef _SHOOTER_H_
#define _SHOOTER_H_
#include "WPILib.h"
#include "WeejusCANJaguar.h"
#include "IOMap.h"
#include "DSMap.h"
#include "MonoEncoder.h"
#include "UDPLogger.h"

#define SHOOTER_LOG 1

#define PID_F 0.0
#define PID_I 0.0
#define PID_P 1.2
#define PID_D 0.0

#define VALUE_TO_ENCODER_OFFSET -0.05f

//robot 2: 5100.0
//robot 1: 4600 (as tested in queue line)
#define SHOOTER_MAX_RPM 	4600
#define SHOOTER_MAX_RPS		(SHOOTER_MAX_RPM / 60.0)

#define ANGLE_SPEED_BROKEN 0.3f
#define ANGLE_SPEED_MED  0.45f
#define ANGLE_SPEED_FAST 0.9f

// indexer speeds
#define DEF_INDEXER_SPEED_FULL 0.0f
#define DEF_INDEXER_SPEED_EMPTY 0.5f
#define DEF_INDEXER_SPEED_MAN 0.25f

//this is used in when encoder is broken
//tells the loop to wait X number of seconds before shooting

#define SPEED_WAIT_TIME 3.5f

// always wait this long (seconds) for disc to drop into shooter
// before moving to the Feed/2nd sensor. this is needed
// so the blade doesn't jam the disc into the shooter's pusher
#define DISC_DROP_WAIT_TIME		0.5f

// max time we wait for the disc to settle after getting
// to the Feed/2nd sensor. this is so we give the disc the
// chance to get into the shooter in case the disc got
// "pushed in" when blades ride over it
#define SETTLE_WAIT_TIME		0.5f

// amount of time to wait for the guard to be lowered before pushing disc into shooter 
#define GUARD_DOWN_WAIT_TIME	0.3f

// amount of time to leave the pusher out (i.e. time it takes disc to be shot)
#define PUSHER_WAIT_TIME		0.5f

// amount of time to wait for pusher to return (and disc to be gone)
// (i.e. if sensor still blocked after this time, then re-shoot 'same' disc)
#define CLEAR_WAIT_TIME			0.3f

// wait after disc has just been fed to allow
// it to settle before moving the rotor
#define FEED_SETTLE_WAIT_TIME	0.1f

//predefined servo positions for the guard
#define GUARD_UP_POS 0.5f
#define GUARD_DOWN_POS 1.0f

//predefined flapper servo positions
#define SERVO_SHOOT_POS 0.7f
#define SERVO_LOAD_POS	0.48f

//pot angle in DEGREES that is needed for feeder to feed
//this is now in a scale of -1 -> +1 since we are ignoring angle and just sending
//the raw slider value
#define ANGLE_FEED 0.44f

//shooter angle in DEGREES that is needed for climbing
//this is now in a scale of -1 -> +1 since we are ignoring angle and just sending
//the raw slider value
#define ANGLE_CLIMB 0.9f

// this must be at least tolerance+angle_climb, to keep the shooter
// tilt moving until the angle_climb is passed
#define ANGLE_CLIMB_TARGET 0.95f

//absolute limits for the pot on the angle controller
#define ANGLEPOT_MAX 1.2f
#define ANGLEPOT_MIN 0.45f

//dead zone for angle input
//used when broken
#define ANGLE_DEAD_ZONE 0.2f

//+/- tolerance used when checking what angle the shooter is currently at
#define ANGLE_TOLERANCE 0.020f
#define SLOW_ANGLE_TOLERANCE 0.050f

//this is to make sure we are not stuck waiting for the pid to reach the setpoint
//if the pid does not reach the setpoint before X number of seconds is up,
//we will override the PID and assume that we are at speed
#define SPEEDPID_WAIT_OVERRIDE 3.0f

typedef enum ShootState {
	sStart = 0, sSpeedWait, sDropGuard, sShoot, sWaitShot, sWaitClear
};
/*
 Sensors:
 as per wiki:
 http://penfieldrobotics.com/wiki/index.php?title=2013:Shooter_Storage_Subteam
 but what is the second banner sensor for?
 plan not to switch pots in software
 camera (center aligned with shooter, wired to allow 180 deg flip)
 
 Lights:
 OK to feed
 # of discs
 is at wanted angle
 
 Motors/Actuators:
 as per wiki (above)
 
 Broken Switches:
 arm angle pot
 shooter encoder
 everything broken (disable shooter)
 */

class WeejusShooter {
public:
	/*
	 * Method: getInstance();
	 * 
	 * Inputs: Null
	 * Outputs: pointer to the only WeejusShooter object
	 * 
	 * Remarks:
	 * Returns a pointer to the only instance of the WeejusShooter
	 * If the instance has not been created, it will be, otherwise the existing instance will be returned
	 */
	static WeejusShooter* getInstance();

	float linearRangeScale(float v, float min, float max, float newMin,
			float newMax);

	//DO NOT CALL THIS
	~WeejusShooter();

	/*
	 * Method: moveToAngle(float);
	 * 
	 * Inputs:
	 * 			float angle: angle to move the shooter to (range -1.0 -> +1.0)
	 * Outputs: Null
	 * 
	 * Remarks:
	 * moves the shooter to the specified angle
	 * Called by WeejusControls when the slider representing angle moves
	 * -1 will move shooter to lower limit
	 * +1 will move shooter to upper limit
	 * anything in between will automatically be scaled to the range between min and max limit
	 * if setAngleBroken()was passed true: moves the shooter to the specified angle
	 * otherwise, positive angle = move up, negative angle = move down
	 */
	void moveToAngle(float angle);

	/*
	 * Method: getIsAtAngle();
	 * 
	 * Inputs: Null
	 * Outputs: bool representing whether the shooter is at the angle it was told to move to
	 * 
	 * Remarks:
	 * returns whether the shooter is at the angle that it was previously told to move to by moveToAngle()
	 * Called by WeejusControls to provide feedback to the drivers
	 */
	bool getIsAtAngle();

	/*
	 * Method: calcShooterAngle()
	 * 
	 * Inputs: Null
	 * Outputs: float representing the current shooter angle in degrees
	 * 
	 * Remarks:
	 * if the pot is broken, this will return -1
	 * 
	 */
	float calcShooterAngle();

	/*
	 * Method: getClimbReady()
	 * 
	 * Inputs: Null
	 * Outputs: bool representing whether the shooter is in the position needed for climbing
	 * 
	 * Remarks:
	 * if the pot is broken, this will return 1 always
	 * 
	 * 
	 */
	bool getClimbReady();

	/*
	 * Method: setNumDiscs(int);
	 * 
	 * Inputs:
	 * 			int discs: integer representing the number of discs put into the robot before the match
	 * Outputs: Null
	 * 
	 * Remarks:
	 * tells WeejusShooter how many discs start in the robot; needed for count
	 */
	void setNumDiscs(int discs);

	/*
	 * Method: init()
	 * 
	 * Inputs: Null
	 * Outputs: Null
	 * 
	 * Remarks:
	 * will reset any state variables
	 * must be called when first switches into teleoperated mode
	 */
	void initialize(bool autonoumous);

	/*
	 * Method: process();
	 * 
	 * Inputs: Null
	 * Outputs: Null
	 * 
	 * Remarks:
	 * updates all motors and internal status variables
	 * automatically indexes discs, will move to feed position once disc has dropped in shooter
	 * MUST be called once at the END of each autotonomous and tele-op loop to continue operation of the shooter
	 * Note: must be called last because will reset turnIndexer to false at end of each loop
	 */
	void process();

	/************/
	/*AUTONOMOUS*/
	/************/

	/*
	 * Method: shootAll();
	 * 
	 * Inputs: Null
	 * Outputs: Null
	 * 
	 * Remarks:
	 * shoots all discs in the storage
	 * must be called every loop
	 * Called by WeejusAuto during autonomous
	 */
	void shootAll();

	/*
	 * Method: shootDone();
	 * 
	 * Inputs: Null
	 * Outputs:
	 * 			bool representing whether the robot is done shooting all discs in the storage
	 * 
	 * Remarks:
	 * Called by WeejusAuto during autonomous
	 */
	bool shootDone();

	void autoWheelOn(bool on);

	/**************/
	/*TELEOPERATED*/
	/**************/

	/*
	 * Method: getNumDiscs();
	 * 
	 * Inputs: Null 
	 * Outputs: integer representing number of discs currently in storage
	 * 
	 * Remarks:
	 * incremented when discs are fed by human player
	 * decremented when discs are shot
	 * called by WeejusControls to provide feedback to the drivers
	 */
	int getNumDiscs();

	/*
	 * Method: getCanFeed();
	 * 
	 * Inputs: Null
	 * Outputs: bool representing whether the human player can give a disc to the robot
	 * 
	 * Remarks:
	 * Only true if:
	 * 		fan blades are in correct position
	 * 		NOT 4 discs in robot
	 * 		Shooter at correct angle
	 * called by WeejusControls to give feedback to the drivers
	 */
	bool getCanFeed();

	/*
	 * Method: shootOverride();
	 * 
	 * Inputs: override: when true shooter wheel speed up will be overrode
	 * Outputs: Null
	 * 
	 * 
	 * Remarks:
	 * this will override the shooter wheel speed up logic,
	 * when called, the wheel will never stop unless in feed mode
	 * MUST be called ONCE
	 */
	void shootOverride(bool override);

	/*
	 * Method: startShoot();
	 * 
	 * Inputs: Null
	 * Outputs: Null
	 * 
	 * Remarks:
	 * begins shooting (call ONCE)
	 * if setEncoderBroken() true, will ONLY lower guard and flip disc into shooter
	 * 		Shooter wheel will always be on
	 * otherwise:
	 * 		Once called, robot will continue shooting until either all discs are discharged or stopShoot() is called
	 * 		Will automatically index discs for shooting
	 * 		Automatically delays shooting until wheel is sped up
	 * called by WeejusControls when the fire button is first pressed
	 */
	void startShoot();

	/*
	 * Method: stopShoot();
	 * 
	 * Inputs: Null
	 * Outputs: Null
	 * 
	 * Remarks:
	 * stops automatic shooting (call ONCE)
	 * once called, all actions begun and currently in progress by shooter will stop
	 * call startShoot() to resume
	 * Called by WeejusControls when the fire button is released
	 */
	void stopShoot();

	/*
	 * Method: feedMode(bool);
	 * 
	 * Inputs: 
	 * 			on: bool representing whether to put the robot in feed mode
	 * Outputs: Null
	 * 
	 * Remarks:
	 * puts the robot in feed mode
	 * when in feed mode, the indexer will only index once until all discs are fed
	 * if the shooter is not at the feed angle, this will do nothing
	 */
	void feedMode(bool on);

	/****************/
	/*MANUEL METHODS*/
	/****************/

	/*
	 * Method: setAngleBroken(bool);
	 * 
	 * Inputs:
	 * 			bool broken: bool representing whether the angle sensing pot is broken
	 * Outputs: Null
	 * 
	 * Remarks:
	 * if true, angle no longer controlled by PID and moveToAngle() checks for direction to move motor
	 * if false, moveToAngle() checks for angle to move to
	 */
	void setAngleBroken(bool broken);

	/*
	 * Method: setEncoderBroken(bool);
	 * 
	 * Inputs:
	 * 			bool broken: bool representing whether the encoder on the shooter wheel is broken
	 * Outputs: Null
	 * 
	 * Remarks:
	 * if true, motor power will be set directly, NOT controlled by PID loop
	 * called by WeejusControls when a broken switch is flipped
	 */
	void setEncoderBroken(bool broken);

	/*
	 * Method: setStorageBroken(bool);
	 * 
	 * Inputs:
	 * 			bool broken: bool representing whether any storage sensors are broken
	 * Outputs: Null
	 * 
	 * Remarks:
	 * if true, storage must be moved manually with turnIndexer();
	 * otherwise, storage will be moved automatically by WeejusShooter
	 * called by WeejusControls when a broken switch is flipped
	 */
	void setStorageBroken(bool broken);

	/*
	 * Method: setDead(bool);
	 * 
	 * Inputs:
	 * 			bool dead: bool representing whether the entire shooter/storage is dead (aka does not work at all)
	 * Outputs: Null
	 * 
	 * Remarks:
	 * tells WeejusShooter that all sensors & motors on shooter & storage systems are broken
	 * If true, nothing on shooter or storage will work
	 * Called by WeejusControls when a broken switch is flipped
	 */
	void setDead(bool dead);

	/*
	 * Method: setSensorDebug(bool);
	 * 
	 * Inputs:
	 * 			bool debug: bool representing whether debug printf
	 * 						is issued when optical flags are tripped
	 * Outputs: Null
	 * 
	 * Remarks:
	 * if true, print debug
	 * if false, no debug
	 */
	void setSensorDebug(bool debug);

	/*
	 * Method: setSpeed(float);
	 * 
	 * Inputs:
	 * 			float speed: value representing speed to run the motor at (range 0.0->1.0)
	 * Outputs: Null
	 * 
	 * Remarks:
	 * sets speed to run the wheel at in broken mode
	 * does nothing unless setEncoderBroken() was passed true
	 */
	void setSpeed(float power);

	/*
	 * Method: turnIndexer();
	 * 
	 * Inputs: Null
	 * Outputs: Null
	 * 
	 * Remarks:
	 * will turn the indexer in one direction at one static speed
	 * indexer will move only as long as this function is called once per loop
	 * called by WeejusControls when a button is pressed
	 */
	void turnIndexer();

	/*
	 * Method: getDebugLine(int);
	 * 
	 * Inputs:
	 * 			int line: integer representing the debug line to get (range 1->6)
	 * Outputs: char array containing the debug line to print
	 * 
	 * Remarks:
	 * will return the line specified by the number passed to the parameter
	 * only can return 6 debug lines
	 * if line < 1, will return line 1
	 * if line > 6, will return line 6
	 */
	char* getDebugLine(int line);

	/*
	 * Debug Line Explanation
	 * 
	 * 1: Loaded			|Load Align			|Feed Align
	 * 2: Guard Position	|Flapper Position
	 * 3: Can Feed			|Shooting			|Fed
	 * 4: Pot Voltage		|Shooter RPM
	 * 5: Wanted Angle		|Wanted Shoot Speed
	 * 6: Num Discs			|Disc Error Count
	 */

private:
	WeejusShooter();
	void initSensors();
	void initOutputs();
	bool checkShooterAngle(float angle);

	void processAngleController();
	void processIndexLoop();
	void runShooterWheel();
	void stopShooterWheel();
	void processShootLoop();

	static WeejusShooter* _theShooter;

	char _dbgBuf[DEBUG_LINE_LENGTH]; //line length is 21. one more byte for null terminator otherwise wndakl3ER#?REw*altcodes*33jus?DK#R#ADAddnwald
	int _numDiscs;
	int _discErrorCount; //this is used to tell if the given # of discs in robot was incorrect
	//indexer state mahcine will increment every time indexer hits feed align sensor without loading a disc
	//if this gets to 4, we know there are no discs in the robot
	double _waitTimeStart; //this is used in autonomous when waiting for the shooter wheel to speed up
	//but the encoder is broken so we cannot use the pid interface
	//if encoder is not broken, this is used for hte PID timeout override
	double _waitShootStart;
	double _waitIndexStart;
	double _speedWaitTimeOvr;
	double _timeOfLastShot;
	// pid loop vars
	double _pidIntegral;
	double _pidLastError;

	//status variables
	//
	//autonomous
	bool _statShootDone;
	//teleop
	ShootState _shootState;
	bool _reInitializeShootState;
	bool _discInShooter;
	bool _reInitializeIndexState; // used when re-entering Auto or Tele-op
	bool _discInFeeder;
	bool _statShootOverride;
	bool _statAutoOverride;
	bool _statFeedMode;
	bool _statCanFeed;
	bool _statShooting;
	bool _statMoveIndexer; //<- manual ONLY
	bool _sensorDebug;

	//broken switches
	bool _brkAngle;
	bool _brkEncoder;
	bool _brkStorage;
	bool _brkDead;

	//sensors
	//angle controller
	AnalogChannel* _sensAngle;
	MonoEncoder* _sensSpeed;
	//shooter
	DigitalInput* _sensLoaded;
	DigitalInput* _sensAlignLoad;
	DigitalInput* _sensAlignFeed;
	DigitalInput* _sensFed;

	//outputs
	//motors
	WeejusCANJaguar* _mtrAngle;
	WeejusCANJaguar* _mtrShoot;
	WeejusCANJaguar* _mtrIndexer;
	Servo* _srvGuard;
	Servo* _srvShoot;

	UDPLogger* _log;

	//input vars for PID loops
	//all vars here will be used during manual mode as well
	//angle controllers
	float _wntAngle; //wanted angle | degrees
	float _shootSpeed; //wanted shoot speed
};

#endif
